/*
 * FPF implementation.
 * 2013 @ Jaroslaw Sliwinski &  Piotr Wisniewski
 */
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/types.h>

#include "fpf.h"
#include "cf_config.h"

/* global table for all interfaces */
#define CF_TABLE_SIZE 256
struct cf_forwarding_entry cf_entries[CF_TABLE_SIZE];

DEFINE_RWLOCK(fentry_lock);

int cf_create_config(void)
{
	size_t i;
	FPF_DEBUG("FPF forward: creating configuration\n");
	for (i = 0; i < CF_TABLE_SIZE; i++)
		cf_remove_fentry(i);
	return 0;
}

void cf_destroy_config(void)
{
	int i;
	FPF_DEBUG("FPF forward: destroying configuration\n");
	for (i = 0; i < CF_TABLE_SIZE; i++)
		cf_remove_fentry(i);
}

struct cf_forwarding_entry *cf_get_fentry(uint8_t key)
{
	FPF_DEBUG("FPF forward: searching for key=%x\n", key);
	return &cf_entries[key];
}

int cf_create_fentry_ethernet(uint8_t key, char *if_name, uint8_t *dst_ethernet)
{
	struct net_device *dev;
	struct cf_forwarding_entry *fe = &cf_entries[key];
	FPF_DEBUG("FPF forward: adding Ethernet entry %x for interface %s\n", key, if_name);
	cf_remove_fentry(key);
	write_lock(&fentry_lock);
	dev = dev_get_by_name(&init_net, if_name);
	if (!dev)
	{
		FPF_DEBUG("FPF forward: intercace %s not found\n", if_name);
		write_unlock(&fentry_lock);
		return -ENOENT;
	}
	/* set values */
	fe->type = CF_ETHERNET;
	fe->dev = dev;
	memcpy(fe->dst_ethernet, dst_ethernet, ETH_ALEN);
	write_unlock(&fentry_lock);
	return 0;
}

int cf_create_fentry_vlan(uint8_t key, char *if_name, uint8_t *dst_ethernet, uint16_t vlan_tag)
{
	struct net_device *dev;
	struct cf_forwarding_entry *fe = &cf_entries[key];
	FPF_DEBUG("FPF forward: adding VLAN entry %x for interface %s\n", key, if_name);
	cf_remove_fentry(key);
	write_lock(&fentry_lock);
	dev = dev_get_by_name(&init_net, if_name);
	if (!dev)
	{
		FPF_DEBUG("FPF forward: intercace %s not found\n", if_name);
		write_unlock(&fentry_lock);
		return -ENOENT;
	}
	/* set values */
	fe->type = CF_VLAN;
	fe->dev = dev;
	fe->vlan_tag = vlan_tag;
	memcpy(fe->dst_ethernet, dst_ethernet, ETH_ALEN);
	write_unlock(&fentry_lock);
	return 0;
}

int cf_create_fentry_device(uint8_t key, char *if_name)
{
	struct net_device *dev;
	struct cf_forwarding_entry *fe = &cf_entries[key];
	FPF_DEBUG("FPF forward: adding device entry %x for interface %s\n", key, if_name);
	cf_remove_fentry(key);
	write_lock(&fentry_lock);
	dev = dev_get_by_name(&init_net, if_name);
	if (!dev)
	{
		FPF_DEBUG("FPF forward: intercace %s not found\n", if_name);
		write_unlock(&fentry_lock);
		return -ENOENT;
	}
	/* set values */
	fe->type = CF_DEVICE;
	fe->dev = dev;
	write_unlock(&fentry_lock);
	return 0;
}


int cf_remove_fentry(uint8_t key)
{
	struct cf_forwarding_entry *fe = &cf_entries[key];
	write_lock(&fentry_lock);
	if (fe->type == CF_NONE)
	{
		write_unlock(&fentry_lock);
		return -ENOENT;
	}
	switch(fe->type) {
	case CF_ETHERNET:
		dev_put(fe->dev);
		break;
	case CF_VLAN:
		dev_put(fe->dev);
		break;
	case CF_DEVICE:
		dev_put(fe->dev);
		break;
	default:
		break;
	}
	fe->type = CF_NONE;
	fe->byte_count = 0;
	fe->packet_count = 0;
	write_unlock(&fentry_lock);
	return 0;
}
