/*
 * FPF implementation.
 * 2013 @ Jaroslaw Sliwinski  &  Piotr Wisniewski
 */
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <net/netlink.h>
#include <net/genetlink.h>

#include "fpf.h"
#include "cf_config.h"
#include "cf_netlink.h"

struct genl_family cf_genl_family =
{
	.id = GENL_ID_GENERATE,
	.hdrsize = 0,
	.name = "fpf.forward",
	.version = 0,
	.maxattr = CF_ATTR_MAX,
};

int cf_add_entry_ethernet_handler(struct sk_buff *skb, struct genl_info *info);
int cf_add_entry_vlan_handler(struct sk_buff *skb, struct genl_info *info);
int cf_add_entry_device_handler(struct sk_buff *skb, struct genl_info *info);
int cf_remove_entry_handler(struct sk_buff *skb, struct genl_info *info);
int cf_get_entry_handler(struct sk_buff *skb, struct genl_info *info);

struct nla_policy cf_nla_policy[] = {
	[CF_ATTR_KEY] = { .type = NLA_U8},
	[CF_ATTR_DEVICE_NAME] = { .type = NLA_STRING, .len = 16},
	[CF_ATTR_DST_ETHERNET_ADDRESS] = { .type = NLA_UNSPEC, .len = 6},
	[CF_ATTR_VLAN_TAG] = { .type = NLA_U16},
};


struct genl_ops cf_genl_ops[] = {
	{.cmd = CF_CMD_ADD_ENTRY_ETHERNET,
	 .flags = GENL_ADMIN_PERM,
	 .policy = cf_nla_policy,
	 .doit = cf_add_entry_ethernet_handler,
	 .dumpit = NULL },
	{.cmd = CF_CMD_ADD_ENTRY_VLAN,
	 .flags = GENL_ADMIN_PERM,
	 .policy = cf_nla_policy,
	 .doit = cf_add_entry_vlan_handler,
	 .dumpit = NULL },
	{.cmd = CF_CMD_ADD_ENTRY_DEVICE,
	 .flags = GENL_ADMIN_PERM,
	 .policy = cf_nla_policy,
	 .doit = cf_add_entry_device_handler,
	 .dumpit = NULL },
	{.cmd = CF_CMD_REMOVE_ENTRY,
	 .flags = GENL_ADMIN_PERM,
	 .policy = cf_nla_policy,
	 .doit = cf_remove_entry_handler,
	 .dumpit = NULL },	
	{.cmd = CF_CMD_GET_ENTRY,
	 .flags = GENL_ADMIN_PERM,
	 .policy = cf_nla_policy,
	 .doit = cf_get_entry_handler,
	 .dumpit = NULL },
};

int cf_add_entry_ethernet_handler(struct sk_buff *skb, struct genl_info *info)
{
	uint8_t key;
	unsigned char *dst_address, *device_name;
	int result;
	FPF_DEBUG("FPF forward: netlink call 'add_entry_ethernet'\n");
	if (!info->attrs[CF_ATTR_KEY])
	{
		printk(KERN_WARNING "FPF forward: missing key\n");
		return -1;
	}
	key = nla_get_u8(info->attrs[CF_ATTR_KEY]);
	if (!info->attrs[CF_ATTR_DST_ETHERNET_ADDRESS])
	{
		printk(KERN_WARNING "FPF forward: missing destination MAC address\n");
		return -1;
	}
	if (nla_len(info->attrs[CF_ATTR_DST_ETHERNET_ADDRESS]) != 6)
	{
		printk(KERN_WARNING "FPF forward: incorrect destination Ethernet address\n");
		return -1;
	}
	dst_address = nla_data(info->attrs[CF_ATTR_DST_ETHERNET_ADDRESS]);
	if (!info->attrs[CF_ATTR_DEVICE_NAME])
	{
		printk(KERN_WARNING "FPF forward: missing device name\n");
		return -1;
	}
	device_name = nla_data(info->attrs[CF_ATTR_DEVICE_NAME]);
	/* try to insert */
	result = cf_create_fentry_ethernet(key, device_name, dst_address);
	return result;
}


int cf_add_entry_vlan_handler(struct sk_buff *skb, struct genl_info *info)
{
	uint8_t key;
	uint16_t vlan_tag;
	unsigned char *dst_address, *device_name;
	int result;
	FPF_DEBUG("FPF forward: netlink call 'add_entry_vlan'\n");
	if (!info->attrs[CF_ATTR_KEY])
	{
		printk(KERN_WARNING "FPF forward: missing key\n");
		return -1;
	}
	key = nla_get_u8(info->attrs[CF_ATTR_KEY]);
	if (!info->attrs[CF_ATTR_DST_ETHERNET_ADDRESS])
	{
		printk(KERN_WARNING "FPF forward: missing destination MAC address\n");
		return -1;
	}
	if (nla_len(info->attrs[CF_ATTR_DST_ETHERNET_ADDRESS]) != 6)
	{
		printk(KERN_WARNING "FPF forward: incorrect destination Ethernet address\n");
		return -1;
	}
	dst_address = nla_data(info->attrs[CF_ATTR_DST_ETHERNET_ADDRESS]);
	if (!info->attrs[CF_ATTR_VLAN_TAG])
	{
		printk(KERN_WARNING "FPF forward: missing VLAN tag\n");
		return -1;
	}
	vlan_tag = nla_get_u16(info->attrs[CF_ATTR_VLAN_TAG]);
	if (!info->attrs[CF_ATTR_DEVICE_NAME])
	{
		printk(KERN_WARNING "FPF forward: missing device name\n");
		return -1;
	}
	device_name = nla_data(info->attrs[CF_ATTR_DEVICE_NAME]);
	/* try to insert */
	result = cf_create_fentry_vlan(key, device_name, dst_address, vlan_tag);
	return result;
}	    

int cf_add_entry_device_handler(struct sk_buff *skb, struct genl_info *info)
{
	uint8_t key;
	unsigned char *device_name;
	int result;
	FPF_DEBUG("FPF forward: netlink call 'add_entry_device'\n");
	if (!info->attrs[CF_ATTR_KEY])
	{
		printk(KERN_WARNING "FPF forward: missing key\n");
		return -1;
	}
	key = nla_get_u8(info->attrs[CF_ATTR_KEY]);
 	if (!info->attrs[CF_ATTR_DEVICE_NAME])
	{
		printk(KERN_WARNING "FPF forward: missing device name\n");
		return -1;
	}
	device_name = nla_data(info->attrs[CF_ATTR_DEVICE_NAME]);
	/* try to insert */
	result = cf_create_fentry_device(key, device_name);
	return result;
}

int cf_remove_entry_handler(struct sk_buff *skb, struct genl_info *info)
{
       	uint8_t key;
	int result;
	FPF_DEBUG("FPF forward: netlink call 'remove_entry'\n");
	if (!info->attrs[CF_ATTR_KEY])
	{
		printk(KERN_WARNING "FPF forward: missing key\n");
		return -1;
	}
	key = nla_get_u8(info->attrs[CF_ATTR_KEY]);
	/* try to remove */
	result = cf_remove_fentry(key);
	return result;

}

int cf_get_entry_handler(struct sk_buff *skb, struct genl_info *info)
{
	uint8_t key;
	struct cf_forwarding_entry *fe;
	struct sk_buff *msg=NULL;
	void *msg_head;
	int result;
	FPF_DEBUG("FPF forward: netlink call 'get_entry'\n");
	if (!info->attrs[CF_ATTR_KEY])
	{
		printk(KERN_WARNING "FPF forward: missing key\n");
		return -1;
	}
	key = nla_get_u8(info->attrs[CF_ATTR_KEY]);
	/* fetch */
	fe = cf_get_fentry(key);
	/* prepare message */
	msg = nlmsg_new(NLMSG_DEFAULT_SIZE, GFP_KERNEL);
	if (msg == NULL)
	{
		printk(KERN_WARNING "FPF forward: unable to crete memory for response\n");
		result = -ENOMEM;
		return result;
	}
	msg_head = genlmsg_put_reply(msg, info, &cf_genl_family, 0, CF_CMD_GET_ENTRY);
	if (!msg_head)
	{
		printk(KERN_WARNING "FPF forward: unable to crete message header\n");
		goto nla_put_failure;
	}
	if (nla_put_u32(msg, CF_RESPONSE_ATTR_TYPE, fe->type))
	goto nla_put_failure;
	if (nla_put_u64(msg, CF_RESPONSE_ATTR_BYTES, fe->byte_count))
	goto nla_put_failure;
	if (nla_put_u64(msg, CF_RESPONSE_ATTR_PACKETS, fe->packet_count))
	goto nla_put_failure;
	switch (fe->type)
	{
	case CF_NONE:
		break;
	case CF_ETHERNET:
		if (nla_put(msg, CF_RESPONSE_ATTR_DST_ETHERNET_ADDRESS, ETH_ALEN, fe->dst_ethernet))
		goto nla_put_failure;
		if (nla_put_string(msg, CF_RESPONSE_ATTR_DEVICE_NAME, fe->dev->name))
		goto nla_put_failure;
		break;
	case CF_VLAN:
		if (nla_put(msg, CF_RESPONSE_ATTR_DST_ETHERNET_ADDRESS, ETH_ALEN, fe->dst_ethernet))
		goto nla_put_failure;
		if (nla_put_string(msg, CF_RESPONSE_ATTR_DEVICE_NAME, fe->dev->name))
		goto nla_put_failure;
		if (nla_put_u16(msg, CF_RESPONSE_ATTR_VLAN_TAG, fe->vlan_tag))
		goto nla_put_failure;
		break;
	case CF_DEVICE:
		if (nla_put_string(msg, CF_RESPONSE_ATTR_DEVICE_NAME, fe->dev->name))
		goto nla_put_failure;
		break;
	};
	result = genlmsg_end(msg, msg_head);
	if (result < 0)
	{
		printk(KERN_WARNING "FPF forward: unable to close message\n");
		goto nla_put_failure;
	}
	result = genlmsg_reply(msg, info);
	FPF_DEBUG("FPF forward: 'get_entry' finished\n");	
	return result;
nla_put_failure:
	result = -ENOMEM;
	if (msg)
		nlmsg_free(msg);
	FPF_DEBUG("FPF forward: 'get_entry' failed\n");	
	return result;
}



int cf_register_netlink(void)
{
	FPF_DEBUG("FPF forward: registering netlink\n");
	if (genl_register_family_with_ops(&cf_genl_family, cf_genl_ops, ARRAY_SIZE(cf_genl_ops)) < 0)
	{
		printk(KERN_WARNING "FPF forward: unable to register generic netlink family or ops\n");
		return -ENOMEM;
	}

	return 0;
}

void cf_deregister_netlink(void)
{
	FPF_DEBUG("FPF forward: deregistering from netlink\n");
	genl_unregister_family(&cf_genl_family);
}
