/*
 * FPF implementation.
 * 2013 @ Jaroslaw Sliwinski &  Piotr Wisniewski
 */
#include <linux/if_vlan.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <net/ip.h>
#include <net/ipv6.h>

#include "fpf.h"
#include "cf_config.h"
#include "cf_protocol.h"


static int cf_skb_recv(struct sk_buff *skb, struct net_device *dev,
			 struct packet_type *ptype, struct net_device *orig_dev);

static inline void cf_decapsulate(struct sk_buff *skb, struct fpf_header *ch);

static struct packet_type fpf_packet_type = {
	.type = cpu_to_be16(COMET_ETHER_TYPE),
	.func = cf_skb_recv,
};

int cf_register_protocol(void)
{
	printk(KERN_DEBUG "FPF forward: registering protocol\n");
	dev_add_pack(&fpf_packet_type);
	return 0;
}

void cf_deregister_protocol(void)
{
	FPF_DEBUG("FPF forward: deregistering protocol\n");
	dev_remove_pack(&fpf_packet_type);
}


int cf_skb_recv(struct sk_buff *skb, struct net_device *dev,
		  struct packet_type *ptype, struct net_device *orig_dev)
{
	struct fpf_header *ch;
	struct cf_forwarding_entry *fe;
	ch = (struct fpf_header *) skb->data;
	FPF_DEBUG("FPF forward: length=%d protocol=%x keylen=%d index=%d key[index]=%x\n", 
	       skb->len, be16_to_cpu(skb->protocol), ch->length, ch->index, ch->key[ch->index]);
	if ( unlikely(sizeof(*ch) + ch->length > skb->len) ) {
		printk(KERN_WARNING "FPF forward: header is to small\n"); 
		goto error_free;
	}
	if (unlikely(ch->index > ch->length)) {
		printk(KERN_WARNING "FPF forward: index greater than length\n"); 
		goto error_free;
	}
	if (ch->index == ch->length) {
		/* remove header and process as plain IP packet */
		cf_decapsulate(skb, ch);
		return NET_RX_SUCCESS;
	}
	fe = cf_get_fentry( ch->key[ch->index] );
	if (unlikely(fe->type == CF_NONE)) {
		FPF_DEBUG("FPF forward: key not found\n");
		goto error_free;
	}
	ch->index++;
	switch (fe->type) {
	case CF_ETHERNET:
		if ( skb_cow_head(skb, ETH_HLEN) < 0 ) {
			printk(KERN_WARNING "FPF forward: unable to add MAC header\n");
			goto error_free;
		}
		fe->byte_count += skb->len;
		fe->packet_count++;
		/* create header */
		skb_push(skb, ETH_HLEN);
		skb_reset_mac_header(skb);
		memcpy(skb->data, fe->dst_ethernet, ETH_ALEN);
		memcpy(skb->data+ETH_ALEN, fe->dev->dev_addr, ETH_ALEN);
		*(uint16_t *)(skb->data+2*ETH_ALEN) = skb->protocol;
		/* send */
		skb->dev = fe->dev;
		dev_queue_xmit(skb);
		break;
	case CF_VLAN:
		if ( skb_cow_head(skb, ETH_HLEN + VLAN_HLEN) < 0 ) {
			printk(KERN_WARNING "FPF forward: unable to add MAC+VLAN header\n");
			goto error_free;
		}
		fe->byte_count += skb->len;
		fe->packet_count++;
		/* create header: MAC + VLAN */
		skb_push(skb, ETH_HLEN + VLAN_HLEN);
		skb_reset_mac_header(skb);
		memcpy(skb->data, fe->dst_ethernet, ETH_ALEN);
		memcpy(skb->data+ETH_ALEN, fe->dev->dev_addr, ETH_ALEN);
		*(uint16_t *)(skb->data+2*ETH_ALEN) = skb->protocol;
		*(uint16_t *)(skb->data+ETH_HLEN) = fe->vlan_tag;
		/* send */
		skb->dev = fe->dev;
		dev_queue_xmit(skb);
		break;
	case CF_DEVICE:
		/* send */
		skb->dev = fe->dev;
		dev_queue_xmit(skb);
		break;
	default:
		goto error_free;
		break;
	}
	return NET_RX_SUCCESS;
error_free:
	kfree_skb(skb);
	return NET_RX_DROP;
}

static inline void cf_decapsulate(struct sk_buff *skb, struct fpf_header *ch)
{
	skb_pull(skb, ch->length + sizeof(*ch));
	skb_reset_network_header(skb);
	switch (skb->data[0] & 0xf0) {
	case 0x40:
		skb->pkt_type = PACKET_HOST;
		skb->protocol = htons(ETH_P_IP);
		netif_receive_skb(skb);
		break;
	case 0x60:
		skb->pkt_type = PACKET_HOST;
		skb->protocol = htons(ETH_P_IPV6);
		netif_receive_skb(skb);
		break;
	default:
		kfree_skb(skb);
	}
}
