/*
 * FPF implementation.
 * 2013 @ Jaroslaw Sliwinski  &  Piotr Wisniewski
 */
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv6.h>
#include <linux/socket.h>
#include <linux/types.h>
#include <net/dst.h>

#include "fpf.h"
#include "ci_netfilter.h"
#include "ci_store.h"

static unsigned int ci_nf_hook_ipv4(unsigned int hooknum, 
				    struct sk_buff *skb,
				    const struct net_device *in,
				    const struct net_device *out,
				    int (*okfn)(struct sk_buff *));

static unsigned int ci_nf_hook_ipv6(unsigned int hooknum, 
				    struct sk_buff *skb,
				    const struct net_device *in,
				    const struct net_device *out,
				    int (*okfn)(struct sk_buff *));

static struct nf_hook_ops ci_nfh_ipv4_pre = 
{
	.hook = &ci_nf_hook_ipv4,
	.owner = THIS_MODULE, 
	.pf = PF_INET,
	.hooknum = NF_INET_PRE_ROUTING,
	.priority = NF_IP_PRI_FIRST
};

static struct nf_hook_ops ci_nfh_ipv4_local = 
{
	.hook = &ci_nf_hook_ipv4,
	.owner = THIS_MODULE, 
	.pf = PF_INET,
	.hooknum = NF_INET_LOCAL_OUT,
	.priority = NF_IP_PRI_FIRST
};

static struct nf_hook_ops ci_nfh_ipv6_pre = 
{
	.hook = &ci_nf_hook_ipv6,
	.owner = THIS_MODULE, 
	.pf = PF_INET6,
	.hooknum = NF_INET_PRE_ROUTING,
	.priority = NF_IP6_PRI_FIRST
};

static struct nf_hook_ops ci_nfh_ipv6_local =
{
	.hook = &ci_nf_hook_ipv6,
	.owner = THIS_MODULE, 
	.pf = PF_INET6,
	.hooknum = NF_INET_LOCAL_OUT,
	.priority = NF_IP6_PRI_FIRST
};

int ci_register_netfilter(void)
{
	FPF_DEBUG("FPF intercept: registering to netfilter\n");
	if (nf_register_hook(&ci_nfh_ipv4_pre) < 0)
	{
		printk(KERN_WARNING "FPF intercept: unable to create hook for IPv4.\n");
		goto error0;
	}
	if (nf_register_hook(&ci_nfh_ipv4_local) < 0)
	{
		printk(KERN_WARNING "FPF intercept: unable to create hook for IPv4.\n");
		goto error1;
	}
	if (nf_register_hook(&ci_nfh_ipv6_pre))
	{
		printk(KERN_WARNING "FPF intercept: unable to create hook for IPv6.\n");
		goto error2;
	}
	if (nf_register_hook(&ci_nfh_ipv6_local))
	{
		printk(KERN_WARNING "FPF intercept: unable to create hook for IPv6.\n");
		goto error3;
	}
	return 0;
error3:
	nf_unregister_hook(&ci_nfh_ipv6_pre);
error2:
	nf_unregister_hook(&ci_nfh_ipv4_local);
error1:
	nf_unregister_hook(&ci_nfh_ipv4_pre);
error0:
	return -ENOMEM;
}

void ci_deregister_netfilter(void)
{
	FPF_DEBUG( "FPF intercept: deregistering from netfilter\n");
	nf_unregister_hook(&ci_nfh_ipv6_local);
	nf_unregister_hook(&ci_nfh_ipv6_pre);
	nf_unregister_hook(&ci_nfh_ipv4_local);
	nf_unregister_hook(&ci_nfh_ipv4_pre);
}

unsigned char dummy_addr[] = {0x07, 0x00, 0x00, 0x00, 0x00, 0x00};

static unsigned int ci_nf_hook_ipv4(unsigned int hooknum,
				    struct sk_buff *skb,
				    const struct net_device *in,
				    const struct net_device *out,
				    int (*okfn)(struct sk_buff *))
{
	uint32_t id;
	size_t byte_count = skb->len;
	size_t key_length;
	char *key;
	struct fpf_header *ch;
	FPF_DEBUG( "FPF intercept: hook IPv4\n");
	FPF_DEBUG( "FPF intercept: length=%d data[0]=%x protocol=%x\n", 
	       skb->len, skb->data[0], be16_to_cpu(skb->protocol));
	/* non-matching packets are forwarded as they are */
	if (ci_find_entry_ipv4(skb->data, byte_count, &id, &key, &key_length) < 0)
		return NF_ACCEPT;
	if (!skb->dev)
		skb->dev = (struct net_device *) out;		
	FPF_DEBUG( "FPF intercept: match for id=%d\n", id);
	if (unlikely(skb_cow_head(skb, key_length+sizeof(*ch) + ETH_HLEN) < 0) )
	{
		printk(KERN_WARNING "FPF intercept: unable to prepend frame"
		       " with %llu bytes\n",
		       (unsigned long long)(key_length + sizeof(*ch) + ETH_HLEN));
		return NF_DROP;
	}
	/* add FPF header */
	ch = (struct fpf_header *) skb_push(skb, key_length+sizeof(*ch));
	skb_reset_network_header(skb);
	ch->length = key_length;
	ch->index = 0;
	memcpy(ch->key, key, key_length);
	skb->protocol = htons(COMET_ETHER_TYPE);
	/* add dummy ethernet header */
	skb_reset_mac_header(skb);
	netif_rx(skb);
	return NF_STOLEN;
}

static unsigned int ci_nf_hook_ipv6(unsigned int hooknum,
				    struct sk_buff *skb,
				    const struct net_device *in,
				    const struct net_device *out,
				    int (*okfn)(struct sk_buff *))
{
	uint32_t id;
	size_t byte_count = skb->len;
	size_t key_length;
	char *key;
	struct fpf_header *ch;
	FPF_DEBUG( "FPF intercept: hook IPv6\n");
	FPF_DEBUG( "FPF intercept: length=%d data[0]=%x protocol=%x\n", 
	       skb->len, skb->data[0], be16_to_cpu(skb->protocol));
	/* non-matching packets are forwarded as they are */
	if (ci_find_entry_ipv6(skb->data, byte_count, &id, &key, &key_length) < 0)
		return NF_ACCEPT;
	FPF_DEBUG( "FPF intercept: match for id=%d\n", id);
	if (!skb->dev)
		skb->dev = (struct net_device *) out;
	if ( unlikely(skb_cow_head(skb, key_length+sizeof(*ch) + ETH_HLEN) < 0) )
	{
		printk(KERN_WARNING "FPF intercept: "
		       "unable to prepend frame with %llud bytes\n",
		       (uint64_t) (key_length + sizeof(*ch)));
		return NF_DROP;
	}
	/* add FPF header */
	ch = (struct fpf_header *) skb_push(skb, key_length+sizeof(*ch));
	skb_reset_network_header(skb);
	ch->length = key_length;
	ch->index = 0;
	memcpy(ch->key, key, key_length);
	skb->protocol = cpu_to_be16(COMET_ETHER_TYPE);
	/* add dummy ethernet header */
	skb_reset_mac_header(skb);
	FPF_DEBUG("FPF intercept: sending modified frame...\n");
	netif_rx(skb);
	return NF_STOLEN;
}
