/*
 * FPF implementation.
 * 2013 @ Jaroslaw Sliwinski &  Piotr Wisniewski
 */
#include <linux/kernel.h>
#include <net/netlink.h>
#include <net/genetlink.h>

#include "fpf.h"
#include "ci_netlink.h"
#include "ci_store.h"

struct genl_family ci_genl_family =
{
	.id = GENL_ID_GENERATE,
	.hdrsize = 0,
	.name = "fpf.intercept",
	.version = 0,
	.maxattr = CI_ATTR_MAX,
};

struct nla_policy ci_nla_policy[] = {
	[CI_ATTR_ID] = { .type = NLA_U32},
	[CI_ATTR_TIMEOUT] = { .type = NLA_U32},
	[CI_ATTR_SRC_ADDRESS_IPV4] = { .type = NLA_UNSPEC, .len = 4},
	[CI_ATTR_DST_ADDRESS_IPV4] = { .type = NLA_UNSPEC, .len = 4},
	[CI_ATTR_SRC_ADDRESS_IPV6] = { .type = NLA_UNSPEC, .len = 16},
	[CI_ATTR_DST_ADDRESS_IPV6] = { .type = NLA_UNSPEC, .len = 16},	
	[CI_ATTR_PROTOCOL] = { .type = NLA_U8},
	[CI_ATTR_SRC_PORT] = { .type = NLA_U16},
	[CI_ATTR_DST_PORT] = { .type = NLA_U16},
	[CI_ATTR_KEY] = { .type = NLA_UNSPEC, .len = CI_KEY_SIZE},
	[CI_ATTR_KEY_LENGTH] = { .type = NLA_U8},
};

int ci_add_entry_ipv4_handler(struct sk_buff *skb, struct genl_info *info);
int ci_remove_entry_ipv4_handler(struct sk_buff *skb, struct genl_info *info);
int ci_add_entry_ipv6_handler(struct sk_buff *skb, struct genl_info *info);
int ci_remove_entry_ipv6_handler(struct sk_buff *skb, struct genl_info *info);
int ci_collect_entries_handler(struct sk_buff *skb, struct genl_info *info);

struct genl_ops ci_genl_ops[] = {
	{.cmd = CI_CMD_ADD_ENTRY_IPV4,
	 .flags = GENL_ADMIN_PERM,
	 .policy = ci_nla_policy,
	 .doit = ci_add_entry_ipv4_handler,
	 .dumpit = NULL },
	{.cmd = CI_CMD_REMOVE_ENTRY_IPV4,
	 .flags = GENL_ADMIN_PERM,
	 .policy = ci_nla_policy,
	 .doit = ci_remove_entry_ipv4_handler,
	 .dumpit = NULL },
	{.cmd = CI_CMD_ADD_ENTRY_IPV6,
	 .flags = GENL_ADMIN_PERM,
	 .policy = ci_nla_policy,
	 .doit = ci_add_entry_ipv6_handler,
	 .dumpit = NULL },
	{.cmd = CI_CMD_REMOVE_ENTRY_IPV6,
	 .flags = GENL_ADMIN_PERM,
	 .policy = ci_nla_policy,
	 .doit = ci_remove_entry_ipv6_handler,
	 .dumpit = NULL },	
	{.cmd = CI_CMD_COLLECT_ENTRIES,
	 .flags = GENL_ADMIN_PERM,
	 .policy = ci_nla_policy,
	 .doit = ci_collect_entries_handler,
	 .dumpit = NULL },
};


int ci_add_entry_ipv4_handler(struct sk_buff *skb, struct genl_info *info)
{
	int id, timeout, protocol, src_port, dst_port, key_length;
	unsigned char *src_address, *dst_address, *key;
	int result;
	FPF_DEBUG("FPF intercept: netlink call 'add_entry_ipv4'\n");
	if (!info->attrs[CI_ATTR_ID])
	{
		printk(KERN_WARNING "FPF intercept: missing id\n");
		return -1;
	}
	id = nla_get_u32(info->attrs[CI_ATTR_ID]);
	if (!info->attrs[CI_ATTR_TIMEOUT])
	{
		printk(KERN_WARNING "FPF intercept: missing timeout\n");
		return -1;
	}
	timeout = nla_get_u32(info->attrs[CI_ATTR_TIMEOUT]);
	if (!info->attrs[CI_ATTR_SRC_ADDRESS_IPV4])
	{
		printk(KERN_WARNING "FPF intercept: missing IPv4 source address\n");
		return -1;
	}
	if (nla_len(info->attrs[CI_ATTR_SRC_ADDRESS_IPV4]) != 4)
	{
		printk(KERN_WARNING "FPF intercept: not an IPv4 source address\n");
		return -1;
	}
	src_address = nla_data(info->attrs[CI_ATTR_SRC_ADDRESS_IPV4]);
	if (!info->attrs[CI_ATTR_DST_ADDRESS_IPV4])
	{
		printk(KERN_WARNING "FPF intercept: missing IPv4 destination address\n");
		return -1;
	}
	if (nla_len(info->attrs[CI_ATTR_DST_ADDRESS_IPV4]) != 4)
	{
		printk(KERN_WARNING "FPF intercept: not an IPv4 destination address\n");
		return -1;
	}
	dst_address = nla_data(info->attrs[CI_ATTR_DST_ADDRESS_IPV4]);
	if (!info->attrs[CI_ATTR_PROTOCOL])
	{
		printk(KERN_WARNING "FPF intercept: missing protocol\n");
		return -1;
	}
	protocol = nla_get_u8(info->attrs[CI_ATTR_PROTOCOL]);
	if (!info->attrs[CI_ATTR_SRC_PORT])
	{
		printk(KERN_WARNING "FPF intercept: missing source port\n");
		return -1;
	}
	src_port = nla_get_u16(info->attrs[CI_ATTR_SRC_PORT]);
	if (!info->attrs[CI_ATTR_SRC_PORT])
	{
		printk(KERN_WARNING "FPF intercept: missing destination port\n");
		return -1;
	}
	dst_port = nla_get_u16(info->attrs[CI_ATTR_DST_PORT]);
	if (!info->attrs[CI_ATTR_KEY])
	{
		printk(KERN_WARNING "FPF intercept: missing key\n");
		return -1;
	}
	if (nla_len(info->attrs[CI_ATTR_KEY]) == 0)
	{
		printk(KERN_WARNING "FPF intercept: zero length key\n");
		return -1;
	}
	key = nla_data(info->attrs[CI_ATTR_KEY]);
	if (!info->attrs[CI_ATTR_KEY_LENGTH])
	{
		printk(KERN_WARNING "FPF intercept: missing key length\n");
		return -1;
	}
	key_length = nla_get_u8(info->attrs[CI_ATTR_KEY_LENGTH]);
	if (key_length <=0 || key_length > 128)
	{
		printk(KERN_WARNING "FPF intercept: wrong key length\n");
		return -1;
	}
	result = ci_add_entry_ipv4(id, timeout, src_address, dst_address, protocol,
				   src_port, dst_port, key, key_length);
	return result;
}

int ci_remove_entry_ipv4_handler(struct sk_buff *skb, struct genl_info *info)
{
	int id;
	int result;
	FPF_DEBUG("FPF intercept: netlink call 'remove_entry_ipv4'\n");
	if (!info->attrs[CI_ATTR_ID])
	{
		printk(KERN_WARNING "FPF intercept: missing id\n");
		return -1;
	}
	id = nla_get_u32(info->attrs[CI_ATTR_ID]);
	result = ci_remove_entry_ipv4(id, NULL, NULL);
	return result;
}


int ci_add_entry_ipv6_handler(struct sk_buff *skb, struct genl_info *info)
{
	int id, timeout, protocol, src_port, dst_port, key_length;
	unsigned char *src_address, *dst_address, *key;
	int result;
	FPF_DEBUG("FPF intercept: netlink call 'add_entry_ipv6'\n");
	if (!info->attrs[CI_ATTR_ID])
	{
		printk(KERN_WARNING "FPF intercept: missing id\n");
		return -1;
	}
	id = nla_get_u32(info->attrs[CI_ATTR_ID]);
	if (!info->attrs[CI_ATTR_TIMEOUT])
	{
		printk(KERN_WARNING "FPF intercept: missing timeout\n");
		return -1;
	}
	timeout = nla_get_u32(info->attrs[CI_ATTR_TIMEOUT]);
	if (!info->attrs[CI_ATTR_SRC_ADDRESS_IPV6])
	{
		printk(KERN_WARNING "FPF intercept: missing IPv6 source address\n");
		return -1;
	}
	if (nla_len(info->attrs[CI_ATTR_SRC_ADDRESS_IPV6]) != 16)
	{
		printk(KERN_WARNING "FPF intercept: not an IPv6 source address\n");
		return -1;
	}
	src_address = nla_data(info->attrs[CI_ATTR_SRC_ADDRESS_IPV6]);
	if (!info->attrs[CI_ATTR_DST_ADDRESS_IPV6])
	{
		printk(KERN_WARNING "FPF intercept: missing IPv6 destination address\n");
		return -1;
	}
	if (nla_len(info->attrs[CI_ATTR_DST_ADDRESS_IPV6]) != 16)
	{
		printk(KERN_WARNING "FPF intercept: not an IPv6 destination address\n");
		return -1;
	}
	dst_address = nla_data(info->attrs[CI_ATTR_DST_ADDRESS_IPV6]);
	if (!info->attrs[CI_ATTR_PROTOCOL])
	{
		printk(KERN_WARNING "FPF intercept: missing protocol\n");
		return -1;
	}
	protocol = nla_get_u8(info->attrs[CI_ATTR_PROTOCOL]);
	if (!info->attrs[CI_ATTR_SRC_PORT])
	{
		printk(KERN_WARNING "FPF intercept: missing source port\n");
		return -1;
	}
	src_port = nla_get_u16(info->attrs[CI_ATTR_SRC_PORT]);
	if (!info->attrs[CI_ATTR_SRC_PORT])
	{
		printk(KERN_WARNING "FPF intercept: missing destination port\n");
		return -1;
	}
	dst_port = nla_get_u16(info->attrs[CI_ATTR_DST_PORT]);
	if (!info->attrs[CI_ATTR_KEY])
	{
		printk(KERN_WARNING "FPF intercept: missing key\n");
		return -1;
	}
	if (nla_len(info->attrs[CI_ATTR_KEY]) == 0)
	{
		printk(KERN_WARNING "FPF intercept: zero length key\n");
		return -1;
	}
	key = nla_data(info->attrs[CI_ATTR_KEY]);
	if (!info->attrs[CI_ATTR_KEY_LENGTH])
	{
		printk(KERN_WARNING "FPF intercept: missing key length\n");
		return -1;
	}
	key_length = nla_get_u8(info->attrs[CI_ATTR_KEY_LENGTH]);
	if (key_length <=0 || key_length > 128)
	{
		printk(KERN_WARNING "FPF intercept: wrong key length\n");
		return -1;
	}
	result = ci_add_entry_ipv6(id, timeout, src_address, dst_address, protocol,
				   src_port, dst_port, key, key_length);
	return result;
}

int ci_remove_entry_ipv6_handler(struct sk_buff *skb, struct genl_info *info)
{
	int id;
	int result;
	FPF_DEBUG("FPF intercept: netlink call 'remove_entry_ipv6'\n");
	if (!info->attrs[CI_ATTR_ID])
	{
		printk(KERN_WARNING "FPF intercept: missing id\n");
		return -1;
	}
	id = nla_get_u32(info->attrs[CI_ATTR_ID]);
	result = ci_remove_entry_ipv6(id, NULL, NULL);
	return result;
}

int ci_collect_entries_handler(struct sk_buff *skb, struct genl_info *info)
{
	int result;
	struct sk_buff *msg=NULL;
	void *msg_head;
	struct expired_stream exs_list, *exs;
	struct list_head *pos, *q;
	FPF_DEBUG("FPF intercept: netlink call 'collect_entries'\n");
	ci_fetch_expired_streams(&exs_list);
	list_for_each_safe(pos, q, &(exs_list.list) )
	{
		exs = list_entry(pos, struct expired_stream, list);
		msg = nlmsg_new(NLMSG_DEFAULT_SIZE, GFP_KERNEL);
		if (msg == NULL)
		{
			printk(KERN_WARNING "FPF intercept: unable to crete memory for response\n");
			break;
		}
		msg_head = genlmsg_put_reply(msg, info, &ci_genl_family, 0, CI_CMD_COLLECT_ENTRIES);
		if (!msg_head)
		{
			printk(KERN_WARNING "FPF intercept: unable to crete message header\n");
			result = -ENOMEM;
			goto nla_put_failure;
		}
		if (nla_put_u32(msg, CI_RESPONSE_ATTR_ID, exs->id))
		goto nla_put_failure;
		if (nla_put_u64(msg, CI_RESPONSE_ATTR_BYTES, exs->byte_count))
		goto nla_put_failure;
		if (nla_put_u32(msg, CI_RESPONSE_ATTR_DURATION, exs->duration))
		goto nla_put_failure;
		result = genlmsg_end(msg, msg_head);
		if (result < 0)
		{
			printk(KERN_WARNING "FPF intercept: unable to close message\n");
			goto nla_put_failure;
		}
		result = genlmsg_reply(msg, info);
		if (result < 0)
			goto cleanup;
	}
	/* final message with zeros */
	msg = nlmsg_new(NLMSG_DEFAULT_SIZE, GFP_KERNEL);
	if (msg == NULL)
	{
		printk(KERN_WARNING "FPF intercept: unable to crete memory for response\n");
		result = -ENOMEM;
		goto cleanup;
	}
	msg_head = genlmsg_put_reply(msg, info, &ci_genl_family, 0, CI_CMD_COLLECT_ENTRIES);
	if (!msg_head)
	{
		printk(KERN_WARNING "FPF intercept: unable to crete message header\n");
		result = -ENOMEM;
		goto nla_put_failure;
	}
	if (nla_put_u32(msg, CI_RESPONSE_ATTR_ID, 0))
	goto nla_put_failure;
	if (nla_put_u64(msg, CI_RESPONSE_ATTR_BYTES, 0))
	goto nla_put_failure;
	if (nla_put_u32(msg, CI_RESPONSE_ATTR_DURATION, 0))
	goto nla_put_failure;
	result = genlmsg_end(msg, msg_head);
	if (result < 0)
	{
		printk(KERN_WARNING "FPF intercept: unable to close message\n");
		goto nla_put_failure;
	}
	printk(KERN_WARNING "FPF intercept: sending message\n");
	result = genlmsg_reply(msg, info);
cleanup:
	list_for_each_safe(pos, q, &(exs_list.list) )
	{
		exs = list_entry(pos, struct expired_stream, list);
		kfree(exs);
	}
	return result;
nla_put_failure:
	result = -ENOMEM;
	if (msg)
		nlmsg_free(msg);
	goto cleanup;
}


int ci_register_netlink(void)
{
	FPF_DEBUG("FPF intercept: registering netlink\n");
	if (genl_register_family_with_ops(&ci_genl_family, ci_genl_ops, ARRAY_SIZE(ci_genl_ops)) < 0)
	{
		printk(KERN_WARNING "FPF intercept: unable to register generic netlink family or ops\n");
		return -ENOMEM;
	}
	FPF_DEBUG("FPF intercept: family id %d\n", ci_genl_family.id);
	return 0;
}

void ci_deregister_netlink(void)
{
	FPF_DEBUG("FPF intercept: deregistering from netlink\n");
	genl_unregister_family(&ci_genl_family);
}
