/*
 * FPF implementation.
 * 2013 @ Jaroslaw Sliwinski  &  Piotr Wisniewski
 */


/* Piotrek W. patch: next line added for compapility with openSuse 12.2 */
#include <linux/delay.h>

#include <linux/unistd.h>
#include <linux/hash.h>
#include <linux/kthread.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/string.h>
#include <linux/time.h>
#include <linux/vmalloc.h>

#include "fpf.h"
#include "ci_store.h"
#include "ci_netlink.h"

#define CI_STORE_BITS4 12
#define CI_STORE_BITS6 12

#define CI_STORE_SIZE4 (1 << CI_STORE_BITS4)
#define CI_STORE_SIZE6 (1 << CI_STORE_BITS6)

/* store */
struct entry4
{
	struct list_head list;
	uint32_t id;
	unsigned long timestamp;
	unsigned long timeout;
	unsigned long time_creation;
	uint8_t address[8]; /* source[4], destination[4] - the same as in IPv4 header */
	uint16_t src_port;
	uint16_t dst_port;
	uint8_t protocol;
	size_t key_length;
	uint64_t byte_count;
	char key[CI_KEY_SIZE];
};

struct entry6
{
	struct list_head list;
	uint32_t id;
	unsigned long timestamp;
	unsigned long timeout;
	unsigned long time_creation;
	uint8_t address[32]; /* source[8], destination[8] - the same as in IPv6 header */
	uint16_t src_port;
	uint16_t dst_port;
	uint8_t protocol;
	size_t key_length;
	uint64_t byte_count;
	char key[CI_KEY_SIZE];
};

struct entry4 **data4 = NULL;
struct entry6 **data6 = NULL;

/* locks for filters */
DEFINE_RWLOCK(lock4);
DEFINE_RWLOCK(lock6);

/* timeout watchdog */

const int CI_WATCHDOG_TIMER = 10000;
struct task_struct *ts;

int ci_watchdog(void *data);

struct expired_stream expired_head4;
struct expired_stream expired_head6;

DEFINE_RWLOCK(exlock4);
DEFINE_RWLOCK(exlock6);

int ci_create_store(void)
{
	FPF_DEBUG("FPF intercept: creating store\n");
	data4 = vmalloc(CI_STORE_SIZE4 * sizeof(*data4));
	if (!data4)
	{
		printk(KERN_WARNING "FPF intercept: unable to create store for IPv4 entries\n");
		goto error0;
	}
	memset(data4, 0, CI_STORE_SIZE4 * sizeof(*data4));
	data6 = vmalloc(CI_STORE_SIZE6 * sizeof(*data6));
	if (!data6)
	{
		printk(KERN_WARNING "FPF intercept: unable to create store for IPv6 entries\n");
		goto error1;
	}
	memset(data6, 0, CI_STORE_SIZE6 * sizeof(*data6));
	rwlock_init(&lock4);
	rwlock_init(&lock6);
	ts = kthread_run(ci_watchdog, NULL, "fpf_watchdog");
	if (ts == ERR_PTR(-ENOMEM))
	{
		printk(KERN_WARNING "FPF intercept: unable to start timeout watchdog\n");
		goto error2;
	}
	FPF_DEBUG("FPF intercept: store ready\n");
	return 0;
error2:
	kfree(data6);
error1:
	kfree(data4);
error0:
	return -ENOMEM;
}

void ci_destroy_store(void)
{
	size_t i;
	struct entry4 *e4;
	struct entry6 *e6;
	struct list_head *pos, *q;
	FPF_DEBUG("FPF intercept: destroying store\n");
	kthread_stop(ts);
	if (data6)
	{
		for (i = 0; i < CI_STORE_SIZE6; i++) 
			if (data6[i])
			{
				list_for_each_safe(pos, q, &(data6[i]->list) )
				{
					FPF_DEBUG("FPF intercept: removing entry\n");
					e6 = list_entry(pos, struct entry6, list);
					list_del(pos);
					kfree(e6);
				}
				FPF_DEBUG("FPF intercept: removing entry\n");
				kfree(data6[i]);
				data6[i] = NULL;
			}
		vfree(data6);
	}
	data6 = NULL;
	if (data4)
	{
		for (i = 0; i < CI_STORE_SIZE4; i++) 
			if (data4[i])
			{
				list_for_each_safe(pos, q, &(data4[i]->list) )
				{
					FPF_DEBUG("FPF intercept: removing entry\n");
					e4 = list_entry(pos, struct entry4, list);
					list_del(pos);
					kfree(e4);
				}
				FPF_DEBUG("FPF intercept: removing entry\n");
				kfree(data4[i]);
				data4[i] = NULL;
			}

		vfree(data4);
	}
	data4 = NULL;
	FPF_DEBUG("FPF intercept: store destroyed\n");
}

int ci_add_entry_ipv4(uint32_t id, unsigned long timeout,
		      uint8_t *src_address, uint8_t *dst_address,
		      uint8_t protocol,
		      uint16_t src_port, uint16_t dst_port,
		      char *key, size_t key_length)
{
	uint32_t hash;
	struct entry4 *new_entry;
	FPF_DEBUG("FPF intercept: add entry %d\n", id);
	/* create entry */
	new_entry = kmalloc(GFP_KERNEL, sizeof(struct entry4));
	if (!new_entry)
	{
		printk(KERN_WARNING "FPF intercept: "
		       "not enough memory for new entry (%llu bytes)\n",
		       (uint64_t)sizeof(struct entry4));
		return -ENOMEM;
	}
	new_entry->id = id;
	new_entry->timestamp = get_seconds();
	new_entry->timeout = timeout;
	new_entry->time_creation = get_seconds();
	memcpy( &(new_entry->address[0]), src_address, 4);
	memcpy( &(new_entry->address[4]), dst_address, 4);
	new_entry->src_port = htons(src_port);
	new_entry->dst_port = htons(dst_port);
	new_entry->protocol = protocol;
	new_entry->key_length = key_length;
	new_entry->byte_count = 0;
	memcpy( &(new_entry->key[0]), key, key_length);
	/* get bucket */
	hash  = hash_32(*((int32_t *)&(new_entry->address[0])), CI_STORE_BITS4);
	hash ^= hash_32(*((int32_t *)&(new_entry->address[4])), CI_STORE_BITS4);
	write_lock(&lock4);
	/* initialize new list if needed */
	if (!data4[hash])
	{
		data4[hash] = new_entry;
		INIT_LIST_HEAD(& (new_entry->list) );
	}
	/* otherwise extend old list */
	else
	{
		list_add_tail(&(new_entry->list), &(data4[hash]->list));
	}
	write_unlock(&lock4);
	return 0;
}

int ci_find_entry_ipv4(uint8_t *ip_header, size_t byte_count,
		       uint32_t *id, char **key, size_t *key_length)
{
	uint32_t hash;
	struct entry4 *entry;
	struct list_head *pos;
	uint16_t src_port, dst_port;
	uint8_t protocol;
	/* get bucket */
	hash  = hash_32(*((int32_t *)&(ip_header[12])), CI_STORE_BITS4);
	hash ^= hash_32(*((int32_t *)&(ip_header[16])), CI_STORE_BITS4);
	/* find */
	read_lock(&lock4);
	entry = data4[hash];
	if (!entry)
	{
		read_unlock(&lock4);
		return -ENOENT;
	}
	protocol = ip_header[9];
	src_port = *((uint16_t *)&(ip_header[20]));
	dst_port = *((uint16_t *)&(ip_header[22]));
	if (memcmp(&(ip_header[12]), entry->address, 8) == 0 &&
	    ( entry->protocol == 0 || entry->protocol == protocol) &&
	    ( entry->src_port == 0 || entry->src_port == src_port) &&
	    ( entry->dst_port == 0 || entry->dst_port == dst_port) )
	{
		entry->timestamp = get_seconds();
		entry->byte_count += byte_count;
		*id = entry->id;
		*key = entry->key;
		*key_length = entry->key_length;
		read_unlock(&lock4);
		return 0;
	}
	list_for_each(pos, &(data4[hash]->list))
	{
		entry = list_entry(pos, struct entry4, list);
		if (memcmp(&(ip_header[12]), entry->address, 8) == 0 &&
		    ( entry->protocol == 0 || entry->protocol == protocol) &&
		    ( entry->src_port == 0 || entry->src_port == src_port) &&
		    ( entry->dst_port == 0 || entry->dst_port == dst_port) )
		{
			entry->timestamp = get_seconds();
			entry->byte_count += byte_count;
			*id = entry->id;
			*key = entry->key;
			*key_length = entry->key_length;
			read_unlock(&lock4);
			return 0;
		}
	}
	read_unlock(&lock4);
	return -ENOENT;;
}

int ci_remove_entry_ipv4(uint32_t id,
			 uint8_t *src_address, uint8_t *dst_address)
{
	uint32_t hash, i;
	struct entry4 *entry, *tmp;
	struct list_head *pos;
	FPF_DEBUG("FPF intercept: remove entry IPv4 %d\n", id);
	if (src_address && dst_address) {
		/* get bucket */
		hash  = hash_32(*((int32_t *)&(src_address[0])), CI_STORE_BITS4);
		hash ^= hash_32(*((int32_t *)&(dst_address[0])), CI_STORE_BITS4);
	} 
	else 
	{
		/* ... or search for the bucket linearly */
		int found = 0;
		for (i = 0; i < CI_STORE_SIZE4; i++)
		{
			entry = data4[i];
			if (!entry)
				continue;
			if (entry->id == id)
			{
				hash = i;
				found = 1;
				break;
			}
			list_for_each(pos, &(entry->list))
			{
				entry = list_entry(pos, struct entry4, list);
				if (entry->id == id)
				{
					hash = i;
					found = 1;
					break;
				}
			}
			if (found)
				break;
		}
		if (!found)
			return -ENOENT;
	}
	/* find */
	write_lock(&lock4);
	entry = data4[hash];
	if (!entry)
	{
		write_unlock(&lock4);
		return -ENOENT;
	}
	/* removal of the head of the list */
	if (entry->id == id)
	{
		if (list_empty( &(entry->list) ) )
		{
			data4[hash] = NULL;
		}
		else
		{
			tmp = list_first_entry( &(entry->list), struct entry4, list);
			list_del( &(entry->list) );
			data4[hash] = tmp;
		}
		kfree(entry);
		write_unlock(&lock4);
		return 0;
	}
	/* removal from the middle of the list */
	list_for_each(pos, &(data4[hash]->list))
	{
		entry = list_entry(pos, struct entry4, list);
		if (entry->id == id)
		{
			list_del( &(entry->list) );
			kfree(entry);
			write_unlock(&lock4);
			return 0;
		}
	}
	write_unlock(&lock4);
	return -ENOENT;;
}


int ci_process_timeout_ipv4(int initialize, uint32_t *id, uint64_t *byte_count, uint32_t *duration,
			    uint8_t *src_address, uint8_t *dst_address)
{
	static size_t i = 0;
	unsigned long now;
	struct entry4 *entry;
	struct list_head *pos;
	if (initialize)
		i = 0;
	now = get_seconds();
	for (; i < CI_STORE_SIZE4; ++i)
	{
		read_lock(&lock4);
		entry = data4[i];
		if (!entry)
		{
			read_unlock(&lock4);
			continue;
		}
		if (now > entry->timestamp + entry->timeout)
		{
			*id = entry->id;
			*byte_count = entry->byte_count;
			*duration = get_seconds() - entry->time_creation;
			memcpy(src_address, &(entry->address[0]), 4);
			memcpy(dst_address, &(entry->address[4]), 4);
			read_unlock(&lock4);
			return 0;
		}
		list_for_each(pos, &(data4[i]->list))
		{
			entry = list_entry(pos, struct entry4, list);
			if (now > entry->timestamp + entry->timeout)
			{
				*id = entry->id;
				*byte_count = entry->byte_count;
				*duration = get_seconds() - entry->time_creation;
				memcpy(src_address, &(entry->address[0]), 4);
				memcpy(dst_address, &(entry->address[4]), 4);
				read_unlock(&lock4);
				return 0;
			}
		}
		read_unlock(&lock4);
	}
	return -ENOENT;
}

/*
 *  IPv6 part.
 */

int ci_add_entry_ipv6(uint32_t id, unsigned long timeout,
		      uint8_t *src_address, uint8_t *dst_address,
		      uint8_t protocol,
		      uint16_t src_port, uint16_t dst_port,
		      char *key, size_t key_length)
{
	uint32_t hash;
	struct entry6 *new_entry;
	FPF_DEBUG("FPF intercept: add entry %d\n", id);
	/* create entry */
	new_entry = kmalloc(GFP_KERNEL, sizeof(struct entry6));
	if (!new_entry)
	{
		printk(KERN_WARNING "FPF intercept: "
		       "not enough memory for new entry (%llu bytes)\n",
		       (uint64_t) sizeof(struct entry6));
		return -ENOMEM;
	}
	new_entry->id = id;
	new_entry->timestamp = get_seconds();
	new_entry->timeout = timeout;
	new_entry->time_creation = get_seconds();
	memcpy( &(new_entry->address[0]), src_address, 16);
	memcpy( &(new_entry->address[16]), dst_address, 16);
	new_entry->src_port = htons(src_port);
	new_entry->dst_port = htons(dst_port);
	new_entry->protocol = protocol;
	new_entry->key_length = key_length;
	new_entry->byte_count = 0;
	memcpy( &(new_entry->key[0]), key, key_length);
	/* get bucket */
	hash  = hash_32(*((int32_t *)&(new_entry->address[0])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(new_entry->address[4])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(new_entry->address[8])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(new_entry->address[12])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(new_entry->address[16])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(new_entry->address[20])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(new_entry->address[24])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(new_entry->address[28])), CI_STORE_BITS6);
	write_lock(&lock6);
	/* initialize new list if needed */
	if (!data6[hash])
	{
		data6[hash] = new_entry;
		INIT_LIST_HEAD(& (new_entry->list) );
	}
	/* otherwise extend old list */
	else
	{
		list_add_tail(&(new_entry->list), &(data6[hash]->list));
	}
	write_unlock(&lock6);
	return 0;
}

int ci_find_entry_ipv6(uint8_t *ip_header, size_t byte_count,
		       uint32_t *id, char **key, size_t *key_length)
{
	uint32_t hash;
	struct entry6 *entry;
	struct list_head *pos;
	uint16_t src_port, dst_port;
	uint8_t protocol;
	/* get bucket */
	hash  = hash_32(*((int32_t *)&(ip_header[8])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(ip_header[12])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(ip_header[16])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(ip_header[20])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(ip_header[24])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(ip_header[28])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(ip_header[32])), CI_STORE_BITS6);
	hash ^= hash_32(*((int32_t *)&(ip_header[36])), CI_STORE_BITS6);
	/* find */
	read_lock(&lock6);
	entry = data6[hash];
	if (!entry)
	{
		read_unlock(&lock6);
		return -ENOENT;
	}
	protocol = ip_header[6];
	src_port = *((uint16_t *)&(ip_header[40]));
	dst_port = *((uint16_t *)&(ip_header[42]));
	if (memcmp(&(ip_header[8]), entry->address, 32) == 0 &&
	    ( entry->protocol == 0 || entry->protocol == protocol) &&
	    ( entry->src_port == 0 || entry->src_port == src_port) &&
	    ( entry->dst_port == 0 || entry->dst_port == dst_port) )
	{
		entry->timestamp = get_seconds();
		entry->byte_count += byte_count;
		*id = entry->id;
		*key = entry->key;
		*key_length = entry->key_length;
		read_unlock(&lock6);
		return 0;
	}
	list_for_each(pos, &(data6[hash]->list))
	{
		entry = list_entry(pos, struct entry6, list);
		if (memcmp(&(ip_header[8]), entry->address, 32) == 0 &&
		    ( entry->protocol == 0 || entry->protocol == protocol) &&
		    ( entry->src_port == 0 || entry->src_port == src_port) &&
		    ( entry->dst_port == 0 || entry->dst_port == dst_port) )
		{
			entry->timestamp = get_seconds();
			entry->byte_count += byte_count;
			*id = entry->id;
			*key = entry->key;
			*key_length = entry->key_length;
			read_unlock(&lock6);
			return 0;
		}
	}
	read_unlock(&lock6);
	return -ENOENT;;
}

int ci_remove_entry_ipv6(uint32_t id,
			 uint8_t *src_address, uint8_t *dst_address)
{
	uint32_t hash, i;
	struct entry6 *entry, *tmp;
	struct list_head *pos;
	FPF_DEBUG("FPF intercept: remove entry IPv6 %d\n", id);
	if (src_address && dst_address) {
		/* get bucket */
		hash  = hash_32(*((int32_t *)&(src_address[0])), CI_STORE_BITS6);
		hash ^= hash_32(*((int32_t *)&(src_address[4])), CI_STORE_BITS6);
		hash ^= hash_32(*((int32_t *)&(src_address[8])), CI_STORE_BITS6);
		hash ^= hash_32(*((int32_t *)&(src_address[12])), CI_STORE_BITS6);
		hash ^= hash_32(*((int32_t *)&(dst_address[0])), CI_STORE_BITS6);
		hash ^= hash_32(*((int32_t *)&(dst_address[4])), CI_STORE_BITS6);
		hash ^= hash_32(*((int32_t *)&(dst_address[8])), CI_STORE_BITS6);
		hash ^= hash_32(*((int32_t *)&(dst_address[12])), CI_STORE_BITS6);
	} 
	else 
	{
		/* ... or search for the bucket linearly */
		int found = 0;
		for (i = 0; i < CI_STORE_SIZE6; i++)
		{
			entry = data6[i];
			if (!entry)
				continue;
			if (entry->id == id)
			{
				hash = i;
				found = 1;
				break;
			}
			list_for_each(pos, &(entry->list))
			{
				entry = list_entry(pos, struct entry6, list);
				if (entry->id == id)
				{
					hash = i;
					found = 1;
					break;
				}
			}
			if (found)
				break;
		}
		if (!found)
			return -ENOENT;
	}
	/* find */
	write_lock(&lock6);
	entry = data6[hash];
	if (!entry)
	{
		write_unlock(&lock6);
		return -ENOENT;
	}
	/* removal of the head of the list */
	if (entry->id == id)
	{
		if (list_empty( &(entry->list) ) )
		{
			data6[hash] = NULL;
		}
		else
		{
			tmp = list_first_entry( &(entry->list), struct entry6, list);
			list_del( &(entry->list) );
			data6[hash] = tmp;
		}
		kfree(entry);
		write_unlock(&lock6);
		return 0;
	}
	/* removal from the middle of the list */
	list_for_each(pos, &(data6[hash]->list))
	{
		entry = list_entry(pos, struct entry6, list);
		if (entry->id == id)
		{
			list_del( &(entry->list) );
			kfree(entry);
			write_unlock(&lock6);
			return 0;
		}
	}
	write_unlock(&lock6);
	return -ENOENT;;
}

int ci_process_timeout_ipv6(int initialize, uint32_t *id, uint64_t *byte_count, uint32_t *duration,
			    uint8_t *src_address, uint8_t *dst_address)
{
	static size_t i = 0;
	unsigned long now;
	struct entry6 *entry;
	struct list_head *pos;
	if (initialize)
		i = 0;
	now = get_seconds();
	for (; i < CI_STORE_SIZE6; ++i)
	{
		read_lock(&lock6);
		entry = data6[i];
		if (!entry)
		{
			read_unlock(&lock6);
			continue;
		}
		if (now > entry->timestamp + entry->timeout)
		{
			*id = entry->id;
			*byte_count = entry->byte_count;
			*duration = get_seconds() - entry->time_creation;
			memcpy(src_address, &(entry->address[0]), 16);
			memcpy(dst_address, &(entry->address[16]), 16);
			read_unlock(&lock6);
			return 0;
		}
		list_for_each(pos, &(data6[i]->list))
		{
			entry = list_entry(pos, struct entry6, list);
			if (now > entry->timestamp + entry->timeout)
			{
				*id = entry->id;
				*byte_count = entry->byte_count;
				*duration = get_seconds() - entry->time_creation;
				memcpy(src_address, &(entry->address[0]), 16);
				memcpy(dst_address, &(entry->address[16]), 16);
				read_unlock(&lock6);
				return 0;
			}
		}
		read_unlock(&lock6);
	}
	return -ENOENT;
}

void ci_fetch_expired_streams(struct expired_stream *exs)
{
	INIT_LIST_HEAD( &(exs->list) );
	write_lock(&exlock4);
	list_splice_tail( &(expired_head4.list), &(exs->list) );
	INIT_LIST_HEAD(& (expired_head4.list) );	
	write_unlock(&exlock4);
	write_lock(&exlock6);
	list_splice_tail( &(expired_head6.list), &(exs->list) );
	INIT_LIST_HEAD(& (expired_head6.list) );	
	write_unlock(&exlock6);
}

int ci_watchdog(void *data)
{
	int i = 0;
	int init;
	uint8_t src_address[16];
	uint8_t dst_address[16];
	uint32_t id;
	uint64_t byte_count;
	uint32_t duration;
	struct expired_stream *exs;
	struct list_head *pos, *q;
	FPF_DEBUG("FPF intercept: timeout watchdog is starting\n");
	INIT_LIST_HEAD(& (expired_head4.list) );
	INIT_LIST_HEAD(& (expired_head6.list) );
	while (!kthread_should_stop())
	{
		if (i >= CI_WATCHDOG_TIMER)
		{
			/*FPF_DEBUG("FPF intercept: timeout watchdog is checking...\n"); */
			i = 0;
			/* find IPv4 */
			init = 1;
			while (ci_process_timeout_ipv4(init, &id, &byte_count, &duration,
						       src_address, dst_address) >= 0)
			{
				init = 0;
				FPF_DEBUG("FPF intercept: IPv4 stream expired id=%d\n", id);
				/* add id to the list for retrieval */
				exs = kmalloc(GFP_KERNEL, sizeof(struct expired_stream));
				write_lock(&exlock4);
				if (exs)
				{
					exs->id = id;
					exs->byte_count = byte_count;
					exs->duration = duration;
					list_add_tail(&(exs->list), &(expired_head4.list));
				}
				else
				{
					printk(KERN_WARNING
					       "FPF intercept: unable to store data for expired stream...\n");
				}
				write_unlock(&exlock4);
				ci_remove_entry_ipv4(id, src_address, dst_address);
			}
			init = 1;
			while (ci_process_timeout_ipv6(init, &id, &byte_count, &duration,
						       src_address, dst_address) >= 0)
			{
				init = 0;
				FPF_DEBUG("FPF intercept: IPv6 stream expired id=%d\n", id);
				/* add id to the list for retrieval */
				exs = kmalloc(GFP_KERNEL, sizeof(struct expired_stream));
				write_lock(&exlock6);
				if (exs)
				{
					exs->id = id;
					exs->byte_count = byte_count;
					exs->duration = duration;
					list_add_tail(&(exs->list), &(expired_head6.list));
				}
				else
				{
					printk(KERN_WARNING
					       "FPF intercept: unable to store data for expired stream...\n");
				}
				write_unlock(&exlock6);
				ci_remove_entry_ipv6(id, src_address, dst_address);
			}
		}
		msleep(1000);
		i += 1000;
	}
	write_lock(&exlock4);
	list_for_each_safe(pos, q, &(expired_head4.list) )
	{
		exs = list_entry(pos, struct expired_stream, list);
		list_del(pos);
		kfree(exs);
	}
	write_unlock(&exlock4);
	write_lock(&exlock6);
	list_for_each_safe(pos, q, &(expired_head6.list) )
	{
		exs = list_entry(pos, struct expired_stream, list);
		list_del(pos);
		kfree(exs);
	}
	write_unlock(&exlock6);
	FPF_DEBUG("FPF intercept: timeout watchdog is stopping\n");
	return 0;
}
