/*
 * FPF implementation.
 * 2013 @ Jaroslaw Sliwinski & Piotr Wisniewski 
 */
#include <ctype.h>
#include <netlink/netlink.h>
#include <netlink/genl/genl.h>
#include <netlink/genl/ctrl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <cf_netlink.h>

#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))

struct nla_policy cf_response_nla_policy[] = {
	[CF_RESPONSE_ATTR_TYPE] = { .type = NLA_U32},
	[CF_RESPONSE_ATTR_DEVICE_NAME] = { .type = NLA_STRING, .maxlen=16},
	[CF_RESPONSE_ATTR_DST_ETHERNET_ADDRESS] = { .type = NLA_UNSPEC, .minlen=6, .maxlen=6},
	[CF_RESPONSE_ATTR_VLAN_TAG] = { .type = NLA_U16},
	[CF_RESPONSE_ATTR_BYTES] = { .type = NLA_U32},
	[CF_RESPONSE_ATTR_PACKETS] = { .type = NLA_U32},
};


int add_entry_ethernet(int key, const char *device, const unsigned char *dst_address)
{
	int result = -1;
	int family;
	struct nl_handle *sock;
	struct nl_msg *msg;
	if (key > 255 || key < 0)
	{
		fprintf(stderr, "Error: key should be in range <00,ff>\n");
		return -1;
	}	
	sock = nl_handle_alloc();
	if (!sock) {
		fprintf(stderr, "Error: Unable to allocate netlink handle\n");
		return -1;
	}
	if (genl_connect(sock) < 0 ) {
		fprintf(stderr, "Error: genl_connect() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	family = genl_ctrl_resolve(sock, "fpf.forward");
	if (family < 0) {
		fprintf(stderr, "Error: genl_ctrl_resolve() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	/* create message */
	msg = nlmsg_alloc();
	if (!msg) {
		fprintf(stderr, "Error: nlmsg_alloc() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	if (genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ,
			family, 0, NLM_F_ACK, CF_CMD_ADD_ENTRY_ETHERNET, 0) == NULL)
	{
		fprintf(stderr, "Error: genlmsg_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u8(msg, CF_ATTR_KEY, key) < 0)
	{
		fprintf(stderr, "Error: nla_put_u8() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_string(msg, CF_ATTR_DEVICE_NAME, device) < 0)
	{
		fprintf(stderr, "Error: nla_put_string() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put(msg, CF_ATTR_DST_ETHERNET_ADDRESS, ETH_ALEN, dst_address) < 0)
	{
		fprintf(stderr, "Error: nla_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
       	}
	if (nl_send_auto_complete(sock, msg) < 0 )
	{
		fprintf(stderr, "Error: nl_send_auto_complete() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	nlmsg_free(msg);
	/* wait for response */
	result = nl_wait_for_ack(sock);
	/* clean up */
	nl_handle_destroy(sock);
	return result;
}

int add_entry_vlan(int key, const char *device, const unsigned char *dst_address, int vlan_tag)
{
	int result = -1;
	int family;
	struct nl_handle *sock;
	struct nl_msg *msg;
	if (key > 255 || key < 0)
	{
		fprintf(stderr, "Error: key should be in range <00,ff>\n");
		return -1;
	}
	if (vlan_tag > (1 << 16) || vlan_tag < 0)
	{
		fprintf(stderr, "Error: vlan_tag should be in range <0000,ffff>\n");
		return -1;
	}
	sock = nl_handle_alloc();
	if (!sock) {
		fprintf(stderr, "Error: Unable to allocate netlink handle\n");
		return -1;
	}
	if (genl_connect(sock) < 0 ) {
		fprintf(stderr, "Error: genl_connect() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	family = genl_ctrl_resolve(sock, "fpf.forward");
	if (family < 0) {
		fprintf(stderr, "Error: genl_ctrl_resolve() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	/* create message */
	msg = nlmsg_alloc();
	if (!msg) {
		fprintf(stderr, "Error: nlmsg_alloc() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	if (genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ,
			family, 0, NLM_F_ACK, CF_CMD_ADD_ENTRY_VLAN, 0) == NULL)
	{
		fprintf(stderr, "Error: genlmsg_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u8(msg, CF_ATTR_KEY, key) < 0)
	{
		fprintf(stderr, "Error: nla_put_u8() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_string(msg, CF_ATTR_DEVICE_NAME, device) < 0)
	{
		fprintf(stderr, "Error: nla_put_string() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put(msg, CF_ATTR_DST_ETHERNET_ADDRESS, ETH_ALEN, dst_address) < 0)
	{
		fprintf(stderr, "Error: nla_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
       	}
	if (nla_put_u16(msg, CF_ATTR_VLAN_TAG, vlan_tag) < 0)
	{
		fprintf(stderr, "Error: nla_put_u16() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nl_send_auto_complete(sock, msg) < 0 )
	{
		fprintf(stderr, "Error: nl_send_auto_complete() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	nlmsg_free(msg);
	/* wait for response */
	result = nl_wait_for_ack(sock);
	/* clean up */
	nl_handle_destroy(sock);
	return result;
}


int add_entry_device(int key, const char *device)
{
	int result = -1;
	int family;
	struct nl_handle *sock;
	struct nl_msg *msg;
	if (key > 255 || key < 0)
	{
		fprintf(stderr, "Error: key should be in range <00,ff>\n");
		return -1;
	}
	sock = nl_handle_alloc();
	if (!sock) {
		fprintf(stderr, "Error: Unable to allocate netlink handle\n");
		return -1;
	}
	if (genl_connect(sock) < 0 ) {
		fprintf(stderr, "Error: genl_connect() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	family = genl_ctrl_resolve(sock, "fpf.forward");
	if (family < 0) {
		fprintf(stderr, "Error: genl_ctrl_resolve() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	/* create message */
	msg = nlmsg_alloc();
	if (!msg) {
		fprintf(stderr, "Error: nlmsg_alloc() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	if (genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ,
			family, 0, NLM_F_ACK, CF_CMD_ADD_ENTRY_DEVICE, 0) == NULL)
	{
		fprintf(stderr, "Error: genlmsg_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u8(msg, CF_ATTR_KEY, key) < 0)
	{
		fprintf(stderr, "Error: nla_put_u8() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_string(msg, CF_ATTR_DEVICE_NAME, device) < 0)
	{
		fprintf(stderr, "Error: nla_put_string() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nl_send_auto_complete(sock, msg) < 0 )
	{
		fprintf(stderr, "Error: nl_send_auto_complete() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	nlmsg_free(msg);
	/* wait for response */
	result = nl_wait_for_ack(sock);
	/* clean up */
	nl_handle_destroy(sock);
	return result;
}

int remove_entry(int key)
{
	int result = -1;
	int family;
	struct nl_handle *sock;
	struct nl_msg *msg;
	if (key > 255 || key < 0)
	{
		fprintf(stderr, "Error: key should be in range <00,ff>\n");
		return -1;
	}
	sock = nl_handle_alloc();
	if (!sock) {
		fprintf(stderr, "Error: Unable to allocate netlink handle\n");
		return -1;
	}
	if (genl_connect(sock) < 0 ) {
		fprintf(stderr, "Error: genl_connect() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	family = genl_ctrl_resolve(sock, "fpf.forward");
	if (family < 0) {
		fprintf(stderr, "Error: genl_ctrl_resolve() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	/* create message */
	msg = nlmsg_alloc();
	if (!msg) {
		fprintf(stderr, "Error: nlmsg_alloc() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	if (genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ,
			family, 0, NLM_F_ACK, CF_CMD_REMOVE_ENTRY, 0) == NULL)
	{
		fprintf(stderr, "Error: genlmsg_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u8(msg, CF_ATTR_KEY, key) < 0)
	{
		fprintf(stderr, "Error: nla_put_u8() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nl_send_auto_complete(sock, msg) < 0 )
	{
		fprintf(stderr, "Error: nl_send_auto_complete() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	nlmsg_free(msg);
	/* wait for response */
	result = nl_wait_for_ack(sock);
	/* clean up */
	nl_handle_destroy(sock);
	return result;
}


int get_entry(int key)
{
	int result = -1, family;
	struct nl_handle *sock;
	struct nl_msg *msg;
	struct sockaddr_nl peer;
	/* response part */
	struct nlmsghdr *nlh;
	struct nlattr *attrs[CF_RESPONSE_ATTR_MAX+1];
	int type;
	uint64_t byte_count;
	uint64_t packet_count;
	char device_name[16];
	char dst_address[6];
	uint16_t vlan_tag;
	unsigned char *mp;
	size_t i;
	if (key > 255 || key < 0)
	{
		fprintf(stderr, "Error: key should be in range <00,ff>\n");
		return -1;
	}
	sock = nl_handle_alloc();
	if (!sock) {
		fprintf(stderr, "Error: Unable to allocate netlink handle\n");
		return -1;
	}
	if (genl_connect(sock) < 0 ) {
		fprintf(stderr, "Error: genl_connect() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	family = genl_ctrl_resolve(sock, "fpf.forward");
	if (family < 0) {
		fprintf(stderr, "Error: genl_ctrl_resolve() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	/* create message */
	msg = nlmsg_alloc();
	if (!msg) {
		fprintf(stderr, "Error: nlmsg_alloc() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	if (genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ,
			family, 0, NLM_F_ACK, CF_CMD_GET_ENTRY, 0) == NULL)
	{
		fprintf(stderr, "Error: genlmsg_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u8(msg, CF_ATTR_KEY, key) < 0)
	{
		fprintf(stderr, "Error: nla_put_u8() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nl_send_auto_complete(sock, msg) < 0 )
	{
		fprintf(stderr, "Error: nl_send_auto_complete() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	nlmsg_free(msg);
	msg = NULL;
	/* read response */
	i = nl_recv(sock, &peer, &mp, NULL);
	if (i <= 0)
	{
		fprintf(stderr, "Error: nl_recv() failure\n");
		goto cleanup;
	}
	nlh = (struct nlmsghdr *) mp;
	i = genlmsg_parse(nlh, 0, attrs, CF_RESPONSE_ATTR_MAX, cf_response_nla_policy);
	if (i < 0)
	{
		fprintf(stderr, "Error: Failed in parsing the response\n");
		goto cleanup;
	}	
	if (!attrs[CF_RESPONSE_ATTR_TYPE])
	{
		fprintf(stderr, "Error: Missing type in response for get entry\n");
		goto cleanup;
	}
	type = nla_get_u32(attrs[CF_RESPONSE_ATTR_TYPE]);
	switch (type) 
	{
	case CF_NONE:
		printf("key=%02x type=none\n", key);
		break;
	case CF_ETHERNET:
		if (!attrs[CF_RESPONSE_ATTR_BYTES])
		{
			fprintf(stderr, "Error: Missing byte count in response for get entry\n");
			goto cleanup;
		}
		byte_count = nla_get_u64(attrs[CF_RESPONSE_ATTR_BYTES]);
		if (!attrs[CF_RESPONSE_ATTR_PACKETS])
		{
			fprintf(stderr, "Error: Missing packet count in response for get entry\n");
			goto cleanup;
		}
		packet_count = nla_get_u64(attrs[CF_RESPONSE_ATTR_PACKETS]);
		if (!attrs[CF_RESPONSE_ATTR_DEVICE_NAME])
		{
			fprintf(stderr, "Error: Missing device name in response for get entry\n");
			goto cleanup;
		}
		strncpy(device_name, nla_get_string(attrs[CF_RESPONSE_ATTR_DEVICE_NAME]), 16);
		if (!attrs[CF_RESPONSE_ATTR_DST_ETHERNET_ADDRESS])
		{
			fprintf(stderr, "Error: Missing destination Ethrernet address in response for get entry\n");
			goto cleanup;
		}
		memcpy(dst_address, nla_data(attrs[CF_RESPONSE_ATTR_DST_ETHERNET_ADDRESS]), ETH_ALEN);
		printf("key=%02x type=ethernet bytes=%llu packets=%llu dev=%s dst=%02x:%02x:%02x:%02x:%02x:%02x\n",
		       key, (long long unsigned int) byte_count,
		       (long long unsigned int) packet_count, device_name,
		       dst_address[0] & 0xff, dst_address[1] & 0xff, dst_address[2] & 0xff,
		       dst_address[3] & 0xff, dst_address[4] & 0xff, dst_address[5] & 0xff);
		break;
	case CF_VLAN:
		if (!attrs[CF_RESPONSE_ATTR_BYTES])
		{
			fprintf(stderr, "Error: Missing byte count in response for get entry\n");
			goto cleanup;
		}
		byte_count = nla_get_u64(attrs[CF_RESPONSE_ATTR_BYTES]);
		if (!attrs[CF_RESPONSE_ATTR_PACKETS])
		{
			fprintf(stderr, "Error: Missing packet count in response for get entry\n");
			goto cleanup;
		}
		packet_count = nla_get_u64(attrs[CF_RESPONSE_ATTR_PACKETS]);
		if (!attrs[CF_RESPONSE_ATTR_DEVICE_NAME])
		{
			fprintf(stderr, "Error: Missing device name in response for get entry\n");
			goto cleanup;
		}
		strncpy(device_name, nla_get_string(attrs[CF_RESPONSE_ATTR_DEVICE_NAME]), 16);
		if (!attrs[CF_RESPONSE_ATTR_DST_ETHERNET_ADDRESS])
		{
			fprintf(stderr, "Error: Missing destination Ethrernet address in response for get entry\n");
			goto cleanup;
		}
		memcpy(dst_address, nla_data(attrs[CF_RESPONSE_ATTR_DST_ETHERNET_ADDRESS]), ETH_ALEN);
		if (!attrs[CF_RESPONSE_ATTR_VLAN_TAG])
		{
			fprintf(stderr, "Error: Missing VLAN tag in response for get entry\n");
			goto cleanup;
		}
		vlan_tag = nla_get_u16(attrs[CF_RESPONSE_ATTR_VLAN_TAG]);
		printf("key=%02x type=vlan bytes=%llu packets=%llu dev=%s dst=%02x:%02x:%02x:%02x:%02x:%02x vlan=%04x\n",
		       key, (long long unsigned int) byte_count,
		       (long long unsigned int) packet_count, device_name,
		       dst_address[0] & 0xff, dst_address[1] & 0xff, dst_address[2] & 0xff,
		       dst_address[3] & 0xff, dst_address[4] & 0xff, dst_address[5] & 0xff, vlan_tag);
		break;
	case CF_DEVICE:
		if (!attrs[CF_RESPONSE_ATTR_BYTES])
		{
			fprintf(stderr, "Error: Missing byte count in response for get entry\n");
			goto cleanup;
		}
		byte_count = nla_get_u64(attrs[CF_RESPONSE_ATTR_BYTES]);
		if (!attrs[CF_RESPONSE_ATTR_PACKETS])
		{
			fprintf(stderr, "Error: Missing packet count in response for get entry\n");
			goto cleanup;
		}
		packet_count = nla_get_u64(attrs[CF_RESPONSE_ATTR_PACKETS]);
		if (!attrs[CF_RESPONSE_ATTR_DEVICE_NAME])
		{
			fprintf(stderr, "Error: Missing device name in response for get entry\n");
			goto cleanup;
		}
		strncpy(device_name, nla_get_string(attrs[CF_RESPONSE_ATTR_DEVICE_NAME]), 16);
		printf("key=%02x type=device bytes=%llu packets=%llu dev=%s\n",
		       key, (long long unsigned int) byte_count,
		       (long long unsigned int) packet_count, device_name);
		break;
	default:
		printf("%02x unknown\n", key);
	};
	/* clean up */
	result = nl_wait_for_ack(sock);
cleanup:
	if (mp)
		free(mp);
	if (msg)
		nlmsg_free(msg);
	if (sock)
		nl_handle_destroy(sock);
	return result;
}

void usage()
{
	fprintf(stderr, 
		"Usage:\n"
		"  cf_tool add_ethernet key device dst.ethernet.address\n"
		"  cf_tool add_vlan key device dst.ethernet.address vlan_tag\n"
		"  cf_tool add_device key device\n"
		"  cf_tool remove key\n"
		"  cf_tool get key\n"
		"  cf_tool get\n"
		"Examples:\n"
		"  cf_tool add_ethernet a7 eth1 aa:11:b0:00:00:01\n"
		"  cf_tool add_vlan a1 eth1 aa:11:b0:00:00:01 0f01\n"
		"  cf_tool add_device 05 gre1\n"
		"  cf_tool remove a7\n"
		"  cf_tool get a1\n");
}

int main(int argc, const char *const *argv)
{
	unsigned int key;
	uint16_t vlan_tag;
	unsigned int eth_address_wide[6];
	unsigned char eth_address[6];
	size_t i, t;
	unsigned int lo, hi;
	int result;
	/* parse */
	if (argc < 2)
	{
		usage();
		exit(1);
	}
	/* add operation - ethernet */
	if (strcmp(argv[1], "add_ethernet") == 0)
	{
		if (argc != 5)
		{
			usage();
			exit(1);
		}
		t = strlen(argv[2]);
		for (i = 0; i < t; i++)
		{
			if (!isxdigit(argv[2][i]))
			{
				fprintf(stderr, "Error: Invalid values in key: %s\n", argv[2]);
				exit(1);
			}
		}
		if (sscanf(argv[2], "%x", &key) != 1)
		{
			fprintf(stderr, "Error: Invalid key: %s\n", argv[2]);
			exit(1);
		}
		t = strlen(argv[4]);
		for (i = 0; i < t; i++)
		{
			if (!isxdigit(argv[4][i]) && argv[4][i] != ':')
			{
				fprintf(stderr, "Error: Invalid values in ethernet address: %s\n", argv[4]);
				exit(1);
			}
		}
		if (sscanf(argv[4], "%x:%x:%x:%x:%x:%x", 
			   &eth_address_wide[0], &eth_address_wide[1], &eth_address_wide[2],
			   &eth_address_wide[3], &eth_address_wide[4], &eth_address_wide[5]) != 6)
		{
			fprintf(stderr, "Error: Invalid ethernet address: %s\n", argv[4]);
			exit(1);
		}
		for (i = 0; i < ETH_ALEN; i++)
		{
			if (eth_address_wide[i] > 255)
			{
				fprintf(stderr, "Error: Invalid ethernet address: %s\n", argv[4]);
				exit(1);
			}
			eth_address[i] = eth_address_wide[i] & 0xff;
		}
		/* try to add */
		result = add_entry_ethernet(key, argv[3], eth_address);
		if (result)
		{
			fprintf(stderr, "Result: failure while adding, code=%d\n", result);
		}
		exit(result);
	}
	/* add operation - vlan */
	if (strcmp(argv[1], "add_vlan") == 0)
	{
		if (argc != 6)
		{
			usage();
			exit(1);
		}
		t = strlen(argv[2]);
		for (i = 0; i < t; i++)
		{
			if (!isxdigit(argv[2][i]))
			{
				fprintf(stderr, "Error: Invalid values in key: %s\n", argv[2]);
				exit(1);
			}
		}
		if (sscanf(argv[2], "%x", &key) != 1)
		{
			fprintf(stderr, "Error: Invalid key: %s\n", argv[2]);
			exit(1);
		}
		t = strlen(argv[4]);
		for (i = 0; i < t; i++)
		{
			if (!isxdigit(argv[4][i]) && argv[4][i] != ':')
			{
				fprintf(stderr, "Error: Invalid values in ethernet address: %s\n", argv[4]);
				exit(1);
			}
		}
		if (sscanf(argv[4], "%x:%x:%x:%x:%x:%x", 
			   &eth_address_wide[0], &eth_address_wide[1], &eth_address_wide[2],
			   &eth_address_wide[3], &eth_address_wide[4], &eth_address_wide[5]) != 6)
		{
			fprintf(stderr, "Error: Invalid ethernet address: %s\n", argv[4]);
			exit(1);
		}
		for (i = 0; i < ETH_ALEN; i++)
		{
			if (eth_address_wide[i] > 255)
			{
				fprintf(stderr, "Error: Invalid ethernet address: %s\n", argv[4]);
				exit(1);
			}
			eth_address[i] = eth_address_wide[i] & 0xff;
		}
		t = strlen(argv[5]);
		for (i = 0; i < t; i++)
		{
			if (!isxdigit(argv[5][i]))
			{
				fprintf(stderr, "Error: Invalid values in VLAN tag: %s\n", argv[5]);
				exit(1);
			}
		}
		if (sscanf(argv[2], "%hx", &vlan_tag) != 1)
		{
			fprintf(stderr, "Error: Invalid key: %s\n", argv[2]);
			exit(1);
		}
		/* try to add */
		result = add_entry_vlan(key, argv[3], eth_address, vlan_tag);
		if (result)
		{
			fprintf(stderr, "Result: failure while adding, code=%d\n", result);
		}
		exit(result);
	}
	/* add operation - pure device */
	if (strcmp(argv[1], "add_device") == 0)
	{
		if (argc != 4)
		{
			usage();
			exit(1);
		}
		t = strlen(argv[2]);
		for (i = 0; i < t; i++)
		{
			if (!isxdigit(argv[2][i]))
			{
				fprintf(stderr, "Error: Invalid values in key: %s\n", argv[2]);
				exit(1);
			}
		}
		if (sscanf(argv[2], "%x", &key) != 1)
		{
			fprintf(stderr, "Error: Invalid key: %s\n", argv[2]);
			exit(1);
		}
		/* try to add */
		result = add_entry_device(key, argv[3]);
		if (result)
		{
			fprintf(stderr, "Result: failure while adding, code=%d\n", result);
		}
		exit(result);
	}
	/* remove operation - all types */
	if (strcmp(argv[1], "remove") == 0)
	{
		if (argc != 3)
		{
			usage();
			exit(1);
		}
		t = strlen(argv[2]);
		for (i = 0; i < t; i++)
		{
			if (!isxdigit(argv[2][i]))
			{
				fprintf(stderr, "Error: Invalid values in key: %s\n", argv[2]);
				exit(1);
			}
		}
		if (sscanf(argv[2], "%x", &key) != 1)
		{
			fprintf(stderr, "Error: Invalid key: %s\n", argv[2]);
			exit(1);
		}
		/* try to remove */
		result = remove_entry(key);
		if (result)
		{
			fprintf(stderr, "Result: failure while removing, code=%d\n", result);
		}
		exit(result);
	}
	/* get operation - all types */
	if (strcmp(argv[1], "get") == 0)
	{
		if (argc != 2 && argc != 3)
		{
			usage();
			exit(1);
		}
		if (argc == 2)
		{
			lo = 0;
			hi = 256;
		}
		else
		{
			t = strlen(argv[2]);
			for (i = 0; i < t; i++)
			{
				if (!isxdigit(argv[2][i]))
				{
					fprintf(stderr, "Error: Invalid values in key: %s\n", argv[2]);
					exit(1);
				}
			}
			if (sscanf(argv[2], "%x", &lo) != 1)
			{
				fprintf(stderr, "Error: Invalid key: %s\n", argv[2]);
				exit(1);
			}
			hi = lo + 1;
		}
		/* get data */
		/* TODO: add header */
		for (; lo < hi; lo++)
		{
			result = get_entry(lo);
			if (result)
			{
				fprintf(stderr, "Result: failure while reading, code=%d\n", result);
				break;
			}
		}
		exit(result);
	}	
	usage();
	return 1;
}
