/*
 * FPF implementation.
 * 2013 @ Jaroslaw Sliwinski & Piotr Wisniewski
 */
#include <ctype.h>
#include <netlink/netlink.h>
#include <netlink/genl/genl.h>
#include <netlink/genl/ctrl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <ci_netlink.h>

#define ARRAY_SIZE(x) ( (unsigned long)(sizeof(x) / sizeof((x)[0])) )

struct nla_policy ci_response_nla_policy[] = {
	[CI_RESPONSE_ATTR_ID] = { .type = NLA_U32},
	[CI_RESPONSE_ATTR_BYTES] = { .type = NLA_U64},
	[CI_RESPONSE_ATTR_DURATION] = { .type = NLA_U32},
};


int add_entry_ipv4(int id, int timeout, const char *src_address, const char *dst_address, int protocol,
		   int src_port, int dst_port, const unsigned char *key, int key_length);
int add_entry_ipv6(int id, int timeout, const char *src_address, const char *dst_address, int protocol,
		   int src_port, int dst_port, const unsigned char *key, int key_length);

int add_entry(int id, int timeout, const char *src_address, const char *dst_address, int protocol,
	      int src_port, int dst_port, const unsigned char *key, int key_length)
{
	struct nl_addr *nl_addr_src, *nl_addr_dst;
	int result = -1;
	char name1[16], name2[16];
	unsigned char key_buffer[128];
	if (key_length > ARRAY_SIZE(key_buffer))
	{
		fprintf(stderr, "Error: key too long, max = %lu\n",
			ARRAY_SIZE(key_buffer));
		return -1;
	}
	nl_addr_src = nl_addr_parse(src_address, AF_UNSPEC);
	if (!nl_addr_src)
	{
		fprintf(stderr, "Error: Unable to parse source address: %s\n", src_address);
		return -1;
	}
	nl_addr_dst = nl_addr_parse(dst_address, AF_UNSPEC);
	if (!nl_addr_dst)
	{
		fprintf(stderr, "Error: Unable to parse destination address: %s\n", dst_address);
		nl_addr_put(nl_addr_src);
		return -1;
	}
	if (nl_addr_get_family(nl_addr_src) != nl_addr_get_family(nl_addr_src))
	{
		nl_af2str(nl_addr_get_family(nl_addr_src), name1, ARRAY_SIZE(name1));
		nl_af2str(nl_addr_get_family(nl_addr_dst), name2, ARRAY_SIZE(name2));
		printf("Error: Source has different address type than destination: '%s' != '%s'\n",
		       name1, name2 );
		nl_addr_put(nl_addr_src);
		nl_addr_put(nl_addr_dst);
		return -1;
	}
	memset(key_buffer, 0, sizeof(key_buffer));
	memcpy(key_buffer, key, key_length);
	nl_af2str(nl_addr_get_family(nl_addr_src), name1, ARRAY_SIZE(name1));
	switch (nl_addr_get_family(nl_addr_src)) 
	{
	case AF_INET:
		result = add_entry_ipv4(id, timeout, 
					nl_addr_get_binary_addr(nl_addr_src),
					nl_addr_get_binary_addr(nl_addr_dst),
					protocol, src_port, dst_port, key_buffer, key_length);
		break;
	case AF_INET6:
		result = add_entry_ipv6(id, timeout, 
					nl_addr_get_binary_addr(nl_addr_src),
					nl_addr_get_binary_addr(nl_addr_dst),
					protocol, src_port, dst_port, key_buffer, key_length);
		break;
	default:
		nl_af2str(nl_addr_get_family(nl_addr_src), name1, ARRAY_SIZE(name1) );
		printf("Error: Unsupported address type '%s'\n", name1);
		break;
	}
	/* clean up */
	nl_addr_put(nl_addr_src);
	nl_addr_put(nl_addr_dst);
	return result;
}

/* binary form - all parameters mus be parsed and already checked */
int add_entry_ipv4(int id, int timeout, const char *src_address, const char *dst_address, int protocol,
		   int src_port, int dst_port, const unsigned char *key, int key_length)
{
	
	struct nl_handle *sock;
	struct nl_msg *msg;
	int family;
	int result;
	sock = nl_handle_alloc();
	if (!sock) {
		fprintf(stderr, "Error: Unable to allocate netlink handle\n");
		return -1;
	}
	if (genl_connect(sock) < 0 ) {
		fprintf(stderr, "Error: genl_connect() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	family = genl_ctrl_resolve(sock, "fpf.intercept");
	if (family < 0) {
		fprintf(stderr, "Error: genl_ctrl_resolve() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	/* create message */
	msg = nlmsg_alloc();
	if (!msg) {
		fprintf(stderr, "Error: nlmsg_alloc() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	if (genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ,
			family, 0, NLM_F_ACK, CI_CMD_ADD_ENTRY_IPV4, 0) == NULL)
	{
		fprintf(stderr, "Error: genlmsg_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u32(msg, CI_ATTR_ID, id) < 0)
	{
		fprintf(stderr, "Error: nla_put_u32() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u32(msg, CI_ATTR_TIMEOUT, timeout) < 0)
	{
		fprintf(stderr, "Error: nla_put_u32() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put(msg, CI_ATTR_SRC_ADDRESS_IPV4, 4, src_address) < 0)
	{
		fprintf(stderr, "Error: nla_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put(msg, CI_ATTR_DST_ADDRESS_IPV4, 4, dst_address) < 0)
	{
		fprintf(stderr, "Error: nla_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u8(msg, CI_ATTR_PROTOCOL, protocol) < 0)
	{
		fprintf(stderr, "Error: nla_put_u8() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u16(msg, CI_ATTR_SRC_PORT, src_port) < 0)
	{
		fprintf(stderr, "Error: nla_put_u16() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u16(msg, CI_ATTR_DST_PORT, dst_port) < 0)
	{
		fprintf(stderr, "Error: nla_put_u16() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put(msg, CI_ATTR_KEY, CI_KEY_SIZE, key) < 0)
	{
		fprintf(stderr, "Error: nla_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
       	}
	if (nla_put_u8(msg, CI_ATTR_KEY_LENGTH, key_length) < 0)
	{
		fprintf(stderr, "Error: nla_put_u8() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nl_send_auto_complete(sock, msg) < 0 )
	{
		fprintf(stderr, "Error: nl_send_auto_complete() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	nlmsg_free(msg);
	/* wait for response */
	result = nl_wait_for_ack(sock);
	/* clean up */
	nl_handle_destroy(sock);
	return result;
}

/* binary form - all parameters mus be parsed and already checked */
int add_entry_ipv6(int id, int timeout, const char *src_address, const char *dst_address, int protocol,
		   int src_port, int dst_port, const unsigned char *key, int key_length)
{
	
	struct nl_handle *sock;
	struct nl_msg *msg;
	int family;
	int result;
	sock = nl_handle_alloc();
	if (!sock) {
		fprintf(stderr, "Error: Unable to allocate handle\n");
		return -1;
	}
	if (genl_connect(sock) < 0 ) {
		fprintf(stderr, "Error: genl_connect() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	family = genl_ctrl_resolve(sock, "fpf.intercept");
	if (family < 0) {
		fprintf(stderr, "Error: genl_ctrl_resolve() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	/* create message */
	msg = nlmsg_alloc();
	if (!msg) {
		fprintf(stderr, "Error: nlmsg_alloc() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	if (genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ,
			family, 0, NLM_F_ACK, CI_CMD_ADD_ENTRY_IPV6, 0) == NULL)
	{
		fprintf(stderr, "Error: genlmsg_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u32(msg, CI_ATTR_ID, id) < 0)
	{
		fprintf(stderr, "Error: nla_put_u32() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u32(msg, CI_ATTR_TIMEOUT, timeout) < 0)
	{
		fprintf(stderr, "Error: nla_put_u32() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put(msg, CI_ATTR_SRC_ADDRESS_IPV6, 16, src_address) < 0)
	{
		fprintf(stderr, "Error: nla_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put(msg, CI_ATTR_DST_ADDRESS_IPV6, 16, dst_address) < 0)
	{
		fprintf(stderr, "Error: nla_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u8(msg, CI_ATTR_PROTOCOL, protocol) < 0)
	{
		fprintf(stderr, "Error: nla_put_u8() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u16(msg, CI_ATTR_SRC_PORT, src_port) < 0)
	{
		fprintf(stderr, "Error: nla_put_u16() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u16(msg, CI_ATTR_DST_PORT, dst_port) < 0)
	{
		fprintf(stderr, "Error: nla_put_u16() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put(msg, CI_ATTR_KEY, CI_KEY_SIZE, key) < 0)
	{
		fprintf(stderr, "Error: nla_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
       	}
	if (nla_put_u8(msg, CI_ATTR_KEY_LENGTH, key_length) < 0)
	{
		fprintf(stderr, "Error: nla_put_u8() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nl_send_auto_complete(sock, msg) < 0 )
	{
		fprintf(stderr, "Error: nl_send_auto_complete() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	nlmsg_free(msg);
	/* wait for response */
	result = nl_wait_for_ack(sock);
	/* clean up */
	nl_handle_destroy(sock);
	return result;
}


int remove_entry(int id)
{
	struct nl_handle *sock;
	struct nl_msg *msg;
	int family;
	int result;
	sock = nl_handle_alloc();
	if (!sock) {
		fprintf(stderr, "Error: Unable to allocate handle\n");
		return -1;
	}
	if (genl_connect(sock) < 0 ) {
		fprintf(stderr, "Error: genl_connect() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	family = genl_ctrl_resolve(sock, "fpf.intercept");
	if (family < 0) {
		fprintf(stderr, "Error: genl_ctrl_resolve() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	/* create message - try IPv4 first */
	msg = nlmsg_alloc();
	if (!msg) {
		fprintf(stderr, "Error: nlmsg_alloc() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	if (genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ,
			family, 0, NLM_F_ACK, CI_CMD_REMOVE_ENTRY_IPV4, 0) == NULL)
	{
		fprintf(stderr, "Error: genlmsg_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u32(msg, CI_ATTR_ID, id) < 0)
	{
		fprintf(stderr, "Error: nla_put_u32() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nl_send_auto_complete(sock, msg) < 0 )
	{
		fprintf(stderr, "Error: nl_send_auto_complete() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	nlmsg_free(msg);
	/* wait for response */
	result = nl_wait_for_ack(sock);
	/* successful removal of IPv4 */
	if (result == 0)
	{
		/* clean up */
		nl_handle_destroy(sock);
		return result;
	}
	/* else try to remove IPv6 */
	msg = nlmsg_alloc();
	if (!msg) {
		fprintf(stderr, "Error: nlmsg_alloc() failed\n");
		nl_handle_destroy(sock);
		return -1;
	}
	if (genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ,
			family, 0, NLM_F_ACK, CI_CMD_REMOVE_ENTRY_IPV6, 0) == NULL)
	{
		fprintf(stderr, "Error: genlmsg_put() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nla_put_u32(msg, CI_ATTR_ID, id) < 0)
	{
		fprintf(stderr, "Error: nla_put_u32() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	if (nl_send_auto_complete(sock, msg) < 0 )
	{
		fprintf(stderr, "Error: nl_send_auto_complete() failed\n");
		nlmsg_free(msg);
		nl_handle_destroy(sock);
		return -1;
	}
	nlmsg_free(msg);
	/* wait for response */
	result = nl_wait_for_ack(sock);
	/* clean up */
	nl_handle_destroy(sock);
	return result;
}

int collect_entries()
{
	struct nl_handle *sock;
	struct nl_msg *msg = NULL;
	struct sockaddr_nl peer;
	int family, i, done = 0, result;
	unsigned char *mp = NULL;
	uint32_t id;
	uint64_t byte_count;
	uint32_t duration;
	struct nlmsghdr *nlh;
	struct nlattr *attrs[CI_RESPONSE_ATTR_MAX+1];
	result = -1;
	sock = nl_handle_alloc();
	if (!sock) {
		fprintf(stderr, "Error: Unable to allocate handle\n");
		goto cleanup;
	}
	if (genl_connect(sock) < 0 ) {
		fprintf(stderr, "Error: genl_connect() failed\n");
		goto cleanup;
	}
	family = genl_ctrl_resolve(sock, "fpf.intercept");
	if (family < 0) {
		fprintf(stderr, "Error: genl_ctrl_resolve() failed\n");
		goto cleanup;
	}
	/* create message - try IPv4 first */
	msg = nlmsg_alloc();
	if (!msg) {
		fprintf(stderr, "Error: nlmsg_alloc() failed\n");
		goto cleanup;
	}
	if (genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ,
			family, 0, 0, CI_CMD_COLLECT_ENTRIES, 0) == NULL)
	{
		fprintf(stderr, "Error: genlmsg_put() failed\n");
		goto cleanup;
	}
	if (nl_send_auto_complete(sock, msg) < 0 )
	{
		fprintf(stderr, "Error: nl_send_auto_complete() failed\n");
		goto cleanup;
	}
	nlmsg_free(msg);
	msg = NULL;
	/* wait for response */
	while (!done)
	{
		i = nl_recv(sock, &peer, &mp, NULL);
		if (i <= 0)
		{
			fprintf(stderr, "Error: nl_recv() failure\n");
			goto cleanup;
		}
		nlh = (struct nlmsghdr *) mp;
		i = genlmsg_parse(nlh, 0, attrs, CI_RESPONSE_ATTR_MAX, ci_response_nla_policy);
		if (i < 0)
		{
			fprintf(stderr, "Error: Parsing of response failed for collect entries\n");
			goto cleanup;
		}	
		if (!attrs[CI_RESPONSE_ATTR_ID])
		{
			fprintf(stderr, "Error: Missing id in response for collect entries\n");
			goto cleanup;
		}
		id = nla_get_u32(attrs[CI_RESPONSE_ATTR_ID]);
		if (!attrs[CI_RESPONSE_ATTR_BYTES])
		{
			fprintf(stderr, "Error: Missing byte count in response for collect entries\n");
			goto cleanup;
		}
		byte_count = nla_get_u64(attrs[CI_RESPONSE_ATTR_BYTES]);
		if (!attrs[CI_RESPONSE_ATTR_DURATION])
		{
			fprintf(stderr, "Error: Missing duration in response for collect entries\n");
			goto cleanup;
		}
		duration = nla_get_u32(attrs[CI_RESPONSE_ATTR_DURATION]);
		free(mp);
		mp = NULL;
		if (id == 0 && duration == 0 && byte_count == 0)
			break;
		printf("%u %u %llu\n",
		       id, duration, (long long unsigned int) byte_count);
	}
	/* clean up */
	result = nl_wait_for_ack(sock);
cleanup:
	if (mp)
		free(mp);
	if (msg)
		nlmsg_free(msg);
	if (sock)
		nl_handle_destroy(sock);
	return result;
}

void usage()
{
	fprintf(stderr, 
		"Usage:\n"
		"  ci_tool add id timeout src.ip.address dst.ip.address protocol src.port dst.port key\n"
		"  ci_tool remove id\n"
		"  ci_tool collect\n"
		"Examples:\n"
		"  ci_tool add 123 60 1.1.1.2 2.2.2.1 17 0 80 1234567890abcdef\n"
		"  ci_tool remove 123\n"
		"  ci_tool collect\n");
}

int main(int argc, const char *const *argv)
{
	unsigned int id;
	unsigned int timeout;
	unsigned int protocol;
	unsigned int port_src;
	unsigned int port_dst;
	unsigned char key[255];
	unsigned int key_length;
	size_t i, j, t;
	unsigned int k;
	int result;
	/* parse */
	if (argc < 2)
	{
		usage();
		exit(1);
	}
	/* add operation */
	if (strcmp(argv[1], "add") == 0)
	{
		if (argc != 10)
		{
			usage();
			exit(1);
		}
		if (sscanf(argv[2], "%u", &id) != 1)
		{
			fprintf(stderr, "Error: Invalid id: %s\n", argv[2]);
			exit(1);
		}
		if (sscanf(argv[3], "%u", &timeout) != 1)
		{
			fprintf(stderr, "Error: Invalid timeout: %s\n", argv[3]);
			exit(1);
		}
		if (sscanf(argv[6], "%u", &protocol) != 1)
		{
			fprintf(stderr, "Error: Invalid protocol: %s\n", argv[6]);
			exit(1);
		}
		if (sscanf(argv[7], "%u", &port_src) != 1)
		{
			fprintf(stderr, "Error: Invalid source port: %s\n", argv[7]);
			exit(1);
		}
		if (sscanf(argv[8], "%u", &port_dst) != 1)
		{
			fprintf(stderr, "Error: Invalid destination port: %s\n", argv[8]);
			exit(1);
		}
		t = strlen(argv[9]);
		if (t == 0)
		{
			fprintf(stderr, "Error: key must have at least 1 byte\n");
			exit(1);
		}
		for (i = 0; i < t; i++)
		{
			if (!isxdigit(argv[9][i]))
			{
				fprintf(stderr, "Error: key contains non-hex value: %s\n", argv[9]);
				exit(1);
			}
		}
		i = 0;
		j = 0;
		key_length = 0;
		if (t % 2 == 1) /* odd number of characters in key */
		{
			if (sscanf(argv[9], "%1x", &k) != 1)
			{
				fprintf(stderr, "Error: key contains non-hex value: %s\n", argv[9]);
				exit(1);
			}
			key[0] = k & 0xff;
			key_length = 1;
			j = 1;
			i = 1;
		}
		for (; i < t; i += 2, j++)
		{
			if (sscanf(&(argv[9][i]), "%2x", &k) != 1)
			{
				fprintf(stderr, "Error: key contains non-hex value: %s\n", argv[9]);
				exit(1);
			}
			key[j] = k & 0xff;
			key_length++;
		}
		/* try to add */
		result = add_entry(id, timeout, argv[4], argv[5], protocol, port_src, port_dst, key, key_length);
		if (result)
		{
			fprintf(stderr, "Result: failure while adding, code=%d\n", result);
		}
		exit(result);
	}
	if (strcmp(argv[1], "remove") == 0)
	{
		if (argc != 3)
		{
			usage();
			exit(1);
		}
		if (sscanf(argv[2], "%u", &id) != 1)
		{
			fprintf(stderr, "Error: Invalid id: %s\n", argv[2]);
			exit(1);
		}
		/* try to remove */
		result = remove_entry(id);
		if (result)
		{
			fprintf(stderr, "Result: failure while removing, code=%d\n", result);
		}
		exit(result);
	}
	if (strcmp(argv[1], "collect") == 0)
	{
		if (argc != 2)
		{
			usage();
			exit(1);
		}
		/* try to collect */
		result = collect_entries();
		if (result)
		{
			fprintf(stderr, "Result: failure while collecting, code=%d\n", result);
		}
		exit(result);
	}
	usage();
	return 1;
}
